"""Contains the implementations for experimental design methods using Emukit

Implements the supported experimental design methods using the Emukit library.

"""

import numpy as np
from emukit.core.initial_designs import RandomDesign
from emukit.experimental_design import ExperimentalDesignLoop
from emukit.experimental_design.acquisitions import (
	IntegratedVarianceReduction,
	ModelVariance,
)
from matplotlib import pyplot as plt

from ..base import EmukitBase
from ..estimators import EmukitEstimator


class EmukitExperimentalDesign(EmukitBase):
	"""The Emukit experimental design method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		experimental_design_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			return self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				experimental_design_kwargs,
			)

		n_init = self.specification.n_init
		X, Y = self.get_X_Y(n_init, target_function)
		method_kwargs = self.specification.method_kwargs
		estimator = EmukitEstimator(method_kwargs)
		estimator.fit(X, Y)

		acquisition_name = self.specification.method
		acquisitions = dict(
			model_variance=ModelVariance,
			integrated_variance_reduction=IntegratedVarianceReduction,
		)
		acquisition_class = acquisitions.get(acquisition_name, None)
		if acquisition_class is None:
			raise ValueError(
				f"Unsupported emulator acquisition type: {acquisition_name}.",
				f"Supported acquisition types are {', '.join(acquisitions)}",
			)
		acquisition = acquisition_class(model=estimator.emulator)

		design_loop = ExperimentalDesignLoop(
			model=estimator.emulator,
			space=self.parameter_space,
			acquisition=acquisition,
			batch_size=1,
		)
		n_iterations = self.specification.n_iterations
		design_loop.run_loop(target_function, stopping_condition=n_iterations)

		self.emulator = estimator
		self.design_loop = design_loop

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		design = RandomDesign(self.parameter_space)
		n_samples = self.specification.n_samples
		X_sample = design.get_samples(n_samples)

		if self.specification.use_shap and outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-param-importances.png",
			)
			self.calculate_shap_importances(
				X_sample,
				self.emulator,
				self.names,
				self.specification.test_size,
				outfile,
			)

		predicted_mu, predicted_std = self.emulator.predict(X_sample, return_std=True)

		observed_data = self.specification.observed_data
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]

		X = np.arange(0, observed_data.shape[0], 1)
		fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)

		axes[0].plot(X, observed_data)
		axes[0].set_title(f"Observed {output_label}")

		mean_predicted_mu = predicted_mu.mean(axis=0)
		mean_predicted_std = predicted_std.mean(axis=0)
		axes[1].plot(X, mean_predicted_mu)

		for mult, alpha in [(1, 0.6), (2, 0.4), (3, 0.2)]:
			axes[1].fill_between(
				X,
				mean_predicted_mu - mult * mean_predicted_std,
				mean_predicted_mu + mult * mean_predicted_std,
				color="C0",
				alpha=alpha,
			)

		axes[1].set_title(f"Emulated {output_label}")

		self.present_fig(
			fig, outdir, time_now, task, experiment_name, f"emulated-{output_label}"
		)



"""Contains the implementations for the Approximate Bayesian Computation methods

Implements the supported Approximate Bayesian Computation methods.

"""

import importlib
from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .pyabc_wrapper import PyABCApproximateBayesianComputation
from .pymc_wrapper import PyMCApproximateBayesianComputation

TASK = "approximate_bayesian_computation"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	pymc=PyMCApproximateBayesianComputation, pyabc=PyABCApproximateBayesianComputation
)

if importlib.util.find_spec("elfi") is not None:
	from ..experimental.abc.elfi_wrapper import ELFIApproximateBayesianComputation

	IMPLEMENTATIONS["elfi"] = ELFIApproximateBayesianComputation


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for Approximate Bayesian Computation.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for Approximate Bayesian Computation.
	"""
	return IMPLEMENTATIONS


class ApproximateBayesianComputationMethodModel(CalibrationModel):
	"""The Approximate Bayesian Computation method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	n_bootstrap: int = Field(description="The number of bootstrap samples", default=5)
	min_population_size: int = Field(
		description="The minimum population size", default=5
	)
	epsilon: float = Field(
		description="The dissimilarity threshold between observed and simulated data",
		default=0,
	)
	sum_stat: str | Callable = Field(
		description="The summary statistic function for observed and simulated data",
		default="identity",
	)
	distance: str | Callable | None = Field(
		description="The distance function between observed and simulated data",
		default=None,
	)
	quantile: float = Field(
		description="Selection quantile used for the sample acceptance threshold",
		default=0.2,
	)


class ApproximateBayesianComputationMethod(CalibrationMethodBase):
	"""The Approximate Bayesian Computation method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: ApproximateBayesianComputationMethodModel,
		engine: str = "pymc",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""ApproximateBayesianComputationMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (ApproximateBayesianComputationMethodModel): The
				calibration specification.
		    engine (str, optional): The Approximate Bayesian
				Computation backend. Defaults to "pymc".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



from .implementation import (
	SensitivityAnalysisMethod,
	SensitivityAnalysisMethodModel,
	get_implementations,
)

_all__ = [
	SensitivityAnalysisMethod,
	SensitivityAnalysisMethodModel,
	get_implementations,
]



"""Contains base classes for various example models

Abstract base classes are defined for the
example simulation models.

"""

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd


class ExampleModelBase(ABC):
	"""The example simulation model abstract class."""

	def __init__(self) -> None:
		"""ExampleModelBase constructor."""
		super().__init__()
		self.GROUND_TRUTH: dict = {}
		self.OUTPUT_LABELS: list = []

	@abstractmethod
	def get_observed_data(self) -> np.ndarray | pd.DataFrame:
		"""Retrieve observed data.

		Raises:
		    NotImplementedError: Error raised for
				the unimplemented abstract method.

		Returns:
		    np.ndarray | pd.DataFrame: The observed data.
		"""
		raise NotImplementedError("get_observed_data() method not implemented.")

	@abstractmethod
	def simulate(self, parameters: dict) -> np.ndarray | pd.DataFrame:
		"""Run the simulation.

			Args:
		        parameters (dict): The simulation parameters.

		    Raises:
				NotImplementedError: Error raised for the
					unimplemented abstract method.

		Returns:
		    np.ndarray | pd.DataFrame: The simulated data.
		"""
		raise NotImplementedError("simulate() method not implemented.")


class ExampleModelContainer:
	"""Container for example models."""

	def __init__(
		self,
		model: ExampleModelBase,
	):
		"""ExampleModelContainer constructor."""
		self.model = model
		self.observed_data = model.get_observed_data()
		self.ground_truth = model.GROUND_TRUTH
		self.output_labels = model.OUTPUT_LABELS



"""Contains the Emukit base class

The defined base class for the Emukit library.

"""

from collections.abc import Callable

import numpy as np
from emukit.core import ContinuousParameter, DiscreteParameter, ParameterSpace
from emukit.core.initial_designs import RandomDesign

from ..data_model import ParameterDataType
from .calibration_base import CalibrationWorkflowBase


class EmukitBase(CalibrationWorkflowBase):
	"""The Emukit base class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		parameters = []
		self.names = []
		self.data_types = []
		self.bounds = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			bounds = self.get_parameter_bounds(spec)
			self.bounds.append(bounds)
			lower_bound, upper_bound = bounds

			data_type = spec.data_type
			self.data_types.append(data_type)

			if data_type == ParameterDataType.DISCRETE:
				discrete_domain = np.arange(lower_bound, upper_bound + 1)
				parameter = DiscreteParameter(parameter_name, discrete_domain)
			else:
				parameter = ContinuousParameter(
					parameter_name, lower_bound, upper_bound
				)
			parameters.append(parameter)

		self.parameter_space = ParameterSpace(parameters)

	def get_X_Y(
		self, n_init: int, target_function: Callable
	) -> tuple[np.ndarray, np.ndarray]:
		"""Get the X and Y matrices.

		Args:
			n_init (int): The number of samples to take
				from the random design.
			target_function (Callable):
				The simulation function.

		Returns:
			tuple[np.ndarray, np.ndarray]: The X and Y matrices.
		"""
		design = RandomDesign(self.parameter_space)
		X = self.specification.X
		if X is None:
			X = design.get_samples(n_init)

		n_replicates = self.specification.n_replicates
		if n_replicates > 1:
			X = np.repeat(X, n_replicates, axis=0)
			self.rng.shuffle(X)

		Y = self.specification.Y
		if Y is None:
			Y = target_function(X)
		return X, Y

	def sample_parameters(self, n_samples: int) -> np.ndarray:
		"""Get new parameter samples.

		Args:
		    n_samples (int): The number of samples.

		Returns:
		    np.ndarray: The parameter samples.
		"""
		design = RandomDesign(self.parameter_space)
		X = design.get_samples(n_samples)
		return X



"""Contains utilities for calculating discrepancy values.

This module defines utility functions for discrepancies
that can be used to compare simulated and observational data.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable
from pydoc import locate

import numpy as np
import sklearn.metrics as metrics
from scipy.spatial import distance as sp_distance
from scipy.special import kl_div
from scipy.stats import (
	energy_distance,
	multivariate_normal,
	norm,
	poisson,
	wasserstein_distance,
)
from scipy.stats import t as t_dist


class DistanceMetricBase(ABC):
	"""The distance metric abstract class."""

	def __init__(self) -> None:
		"""DistanceMetricBase constructor."""
		super().__init__()

	@abstractmethod
	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.

		Raises:
		    NotImplementedError: Error raised for the
				unimplemented abstract method.
		"""
		raise NotImplementedError("calculate() method not implemented.")


class L1Norm(DistanceMetricBase):
	"""The L1 norm distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = np.linalg.norm(observed - simulated, ord=1)
		return distance


class L2Norm(DistanceMetricBase):
	"""The L2 norm distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = np.linalg.norm(observed - simulated, ord=2)
		return distance


class WassersteinDistance(DistanceMetricBase):
	"""The Wasserstein ID distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = wasserstein_distance(observed, simulated)
		return distance


class KlDivergence(DistanceMetricBase):
	"""The K-L divergence."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = kl_div(observed, simulated).sum()
		return distance


class EnergyDistance(DistanceMetricBase):
	"""The energy distance distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = energy_distance(observed, simulated)
		return distance


class BrayCurtisDistance(DistanceMetricBase):
	"""The Bray-Curtis distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.braycurtis(observed, simulated)
		return distance


class CanberraDistance(DistanceMetricBase):
	"""The Canberra distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.canberra(observed, simulated)
		return distance


class ChebyshevDistance(DistanceMetricBase):
	"""The Chebyshev distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.chebyshev(observed, simulated)
		return distance


class CorrelationDistance(DistanceMetricBase):
	"""The correlation distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.correlation(observed, simulated)
		return distance


class CosineDistance(DistanceMetricBase):
	"""The Cosine distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.cosine(observed, simulated)
		return distance


class MinkowskiDistance(DistanceMetricBase):
	"""The Minkowski distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.minkowski(observed, simulated)
		return distance


class JensenShannonDistance(DistanceMetricBase):
	"""The Jensen-Shannon distance."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		distance = sp_distance.jensenshannon(observed, simulated)
		return distance


class MeanSquaredError(DistanceMetricBase):
	"""The mean squared error distance."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		multioutput: str = "uniform_average",
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			multioutput (str, optional): Defines aggregating of multiple output values.
				Defaults to 'uniform_average'.
		"""
		distance = metrics.mean_squared_error(
			observed, simulated, multioutput=multioutput
		)
		return distance


class MeanAbsoluteError(DistanceMetricBase):
	"""The mean absolute error distance."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		multioutput: str = "uniform_average",
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			multioutput (str, optional): Defines aggregating of multiple output values.
				Defaults to 'uniform_average'.
		"""
		distance = metrics.mean_absolute_error(
			observed, simulated, multioutput=multioutput
		)
		return distance


class RootMeanSquaredError(DistanceMetricBase):
	"""The root mean squared error distance."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		multioutput: str = "uniform_average",
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			multioutput (str, optional): Defines aggregating of multiple output values.
				Defaults to 'uniform_average'.
		"""
		distance = metrics.root_mean_squared_error(
			observed, simulated, multioutput=multioutput
		)
		return distance


class MeanPinballLoss(DistanceMetricBase):
	"""The mean pinball loss distance."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		multioutput: str = "uniform_average",
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			multioutput (str, optional): Defines aggregating of multiple output values.
				Defaults to 'uniform_average'.
		"""
		distance = metrics.mean_pinball_loss(
			observed, simulated, multioutput=multioutput
		)
		return distance


class MeanAbsolutePercentageError(DistanceMetricBase):
	"""The mean absolute percentage error distance."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		multioutput: str = "uniform_average",
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			multioutput (str, optional): Defines aggregating of multiple output values.
				Defaults to 'uniform_average'.
		"""
		distance = metrics.mean_absolute_percentage_error(
			observed, simulated, multioutput=multioutput
		)
		return distance


class MedianAbsoluteError(DistanceMetricBase):
	"""The median absolute error distance."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		multioutput: str = "uniform_average",
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			multioutput (str, optional): Defines aggregating of multiple output values.
				Defaults to 'uniform_average'.
		"""
		distance = metrics.median_absolute_error(
			observed, simulated, multioutput=multioutput
		)
		return distance


def get_distance_metric_func(distance_metric: str) -> Callable:
	"""Get the distance metric function by name.

	Args:
	    distance_metric (str): The distance metric name.

	Returns:
	    Callable: The distance metric function.
	"""
	distance_metric = distance_metric.replace("_", " ").title().replace(" ", "")
	module = "calisim.statistics.discrepancy"
	func: Callable = locate(f"{module}.{distance_metric}")  # type: ignore[assignment]
	return func


class GaussianLogLikelihood(DistanceMetricBase):
	"""The Gaussian log likelihood."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		sigma: float | np.ndarray | None = None,
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			sigma (float | np.ndarray | None, optional): The standard
				deviation. Defaults to None.
		"""
		if sigma is None:
			residuals = observed - simulated
			sigma = np.std(residuals, ddof=1)

		log_likelihood = norm.logpdf(observed, loc=simulated, scale=sigma).sum()
		return log_likelihood


class StudentsTLogLikelihood(DistanceMetricBase):
	"""The Student's t log likelihood."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		nu: int = 1,
		sigma: float | np.ndarray | None = None,
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			nu (int, optional): The degrees of freedom for the
				t distribution. Defaults to 1.
			sigma (float | np.ndarray | None, optional): The standard
				deviation. Defaults to None.
		"""
		if sigma is None:
			residuals = observed - simulated
			sigma = np.std(residuals, ddof=1)

		log_likelihood = t_dist.logpdf(
			observed, df=nu, loc=simulated, scale=sigma
		).sum()
		return log_likelihood


class PoissonLogLikelihood(DistanceMetricBase):
	"""The Poisson log likelihood."""

	def calculate(
		self, observed: np.ndarray, simulated: np.ndarray
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
		"""
		log_likelihood = poisson.logpmf(observed, mu=simulated).sum()
		return log_likelihood


class MultivariateNormalLogLikelihood(DistanceMetricBase):
	"""The multivariate normal log likelihood."""

	def calculate(
		self,
		observed: np.ndarray,
		simulated: np.ndarray,
		cov_matrix: np.ndarray | None = None,
		jitter: float = 1e-7,
	) -> float | np.ndarray:
		"""Calculate the distance between observed and simulated data.

		Args:
		    observed (np.ndarray): The observed data.
		    simulated (np.ndarray): The simulated data.
			cov_matrix (np.ndarray | None, optional): The covariance
				matrix. Defaults to None.
			jitter (float) Jitter to stabilise the inversion of
				the covariance matrix.
		"""
		if cov_matrix is None:
			residuals = observed - simulated
			cov_matrix = np.cov(residuals, rowvar=False, ddof=1)
			cov_matrix = cov_matrix + jitter * np.eye(cov_matrix.shape[0])

		log_likelihoods = [
			multivariate_normal.logpdf(obs, mean=pred, cov=cov_matrix)
			for obs, pred in zip(observed, simulated)
		]

		log_likelihood = np.sum(log_likelihoods)
		return log_likelihood



"""Contains the implementations for simulation-based inference methods using
LAMPE

Implements the supported simulation-based inference methods using
the LAMPE library.

"""

from itertools import islice
from typing import Any

import numpy as np
import pandas as pd
import torch
import torch.distributions as dist
import torch.optim as optim
from lampe.data import JointLoader
from lampe.diagnostics import expected_coverage_mc
from lampe.inference import NPE, NPELoss
from lampe.plots import coverage_plot
from lampe.utils import GDStep
from matplotlib import pyplot as plt
from sbi import analysis as analysis

from ..base import SimulationBasedInferenceBase
from ..data_model import DistributionModel, ParameterDataType, ParameterEstimateModel


class PriorCollection:
	"""A wrapper around a collection of priors."""

	def __init__(self, priors: list[dist.Distribution]) -> None:
		"""PriorCollection constructor.

		Args:
		    priors (list[dist.Distribution]): The list of prior distributions.
		"""
		self.parameters = priors

	def sample(self, batch_shape: tuple = ()) -> torch.Tensor:
		"""Sample from the priors.

		Args:
		    batch_shape (tuple, optional): The batch shape of
				the sampled priors. Defaults to ().

		Returns:
		    torch.Tensor: The sampled priors.
		"""
		prior_sample = []
		for prior in self.parameters:
			prior_sample.append(prior.sample(batch_shape).squeeze())
		return torch.stack(prior_sample).T


class LAMPESimulationBasedInference(SimulationBasedInferenceBase):
	"""The LAMPE simulation-based inference method class."""

	def preprocess(
		self, theta: torch.Tensor, parameter_spec: list[DistributionModel] | Any | None
	) -> torch.Tensor:
		"""Normalise the parameters of the simulation.
		Args:
		    theta (torch.Tensor): The simulation parameters.
		    parameter_spec (list[DistributionModel] | Any | None):
				The parameter specification.
		Raises:
		    ValueError: Error raised when an unsupported distribution is provided.
		Returns:
		    torch.Tensor: The normalised parameters.
		"""
		param_values = []
		for i, spec in enumerate(parameter_spec):  # type: ignore[arg-type]
			x = theta[i]
			distribution_name = spec.distribution_name
			if (
				distribution_name == "uniform"
				or spec.data_type == ParameterDataType.DISCRETE
			):
				lower_bound, upper_bound = self.get_parameter_bounds(spec)
				param_value = 2 * (x - lower_bound) / (upper_bound - lower_bound) - 1
			elif distribution_name == "normal":
				mu, sd = self.get_parameter_bounds(spec)
				param_value = (x - mu) / sd
			else:
				raise ValueError(
					f"Unsupported distribution for LAMPE: {distribution_name}"
				)
			param_values.append(param_value)
		return torch.Tensor(param_values)

	def postprocess(
		self,
		samples: torch.Tensor,
		parameter_spec: list[DistributionModel] | Any | None,
	) -> torch.Tensor:
		"""Reverse normalise the parameters of the simulation.
		Args:
		    samples (torch.Tensor): The normalised parameters.
		    parameter_spec (list[DistributionModel] | Any | None):
				The parameter specification.
		Raises:
		    ValueError: Error raised when an unsupported distribution is provided.
		Returns:
		    torch.Tensor: The denormalised parameters.
		"""
		param_values = []
		for sample in samples:
			norm_param_values = []
			for i, spec in enumerate(parameter_spec):  # type: ignore[arg-type]
				x = sample[i]
				distribution_name = spec.distribution_name
				if (
					distribution_name == "uniform"
					or spec.data_type == ParameterDataType.DISCRETE
				):
					lower_bound, upper_bound = self.get_parameter_bounds(spec)
					param_value = (x + 1) / 2 * (
						upper_bound - lower_bound
					) + lower_bound
				elif distribution_name == "normal":
					mu, sd = self.get_parameter_bounds(spec)
					param_value = x * sd + mu
				else:
					raise ValueError(
						f"Unsupported distribution for LAMPE: {distribution_name}"
					)
				norm_param_values.append(param_value)
			param_values.append(norm_param_values)
		return torch.Tensor(param_values)

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		super().specify()
		self.parameters: PriorCollection = PriorCollection(self.parameters)  # type: ignore[assignment]

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""

		sbi_kwargs = self.get_calibration_func_kwargs()

		def simulator_func(X: np.ndarray) -> np.ndarray:
			X = X.detach().cpu().numpy().T
			X = [X]
			results = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				sbi_kwargs,
			)
			results = results[0]
			return torch.from_numpy(results).float()

		loader = JointLoader(
			self.parameters, simulator_func, batch_size=1, vectorized=True
		)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		estimator = NPE(
			len(self.names),
			len(self.specification.observed_data),  # type: ignore[arg-type]
			**method_kwargs,
		)
		loss = NPELoss(estimator)
		optimizer = optim.Adam(estimator.parameters(), lr=self.specification.lr)
		step = GDStep(optimizer, clip=1.0)
		estimator.train()

		parameter_spec = self.specification.parameter_spec.parameters
		for epoch in range(self.specification.n_iterations):
			for theta, x in islice(loader, self.specification.num_simulations):
				neg_log_p = loss(self.preprocess(theta, parameter_spec), x)
				step(neg_log_p)
			if self.specification.verbose:
				print(f"Epoch {epoch + 1} : Negative log-likelihood {neg_log_p}")  # type: ignore[possibly-undefined]

		self.loader = loader
		self.estimator = estimator
		self.loss = loss
		self.optimizer = optimizer
		self.step = step

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		parameter_spec = self.specification.parameter_spec.parameters
		x_star = torch.from_numpy(self.specification.observed_data).float()
		n_draws = self.specification.n_samples
		with torch.no_grad():
			posterior_samples = self.estimator.flow(x_star).sample((n_draws,))
			posterior_samples = self.postprocess(posterior_samples, parameter_spec)

		for plot_func in [analysis.pairplot, analysis.marginal_plot]:
			plt.rcParams.update({"font.size": 8})
			fig, _ = plot_func(
				posterior_samples, figsize=self.specification.figsize, labels=self.names
			)
			self.present_fig(
				fig,
				outdir,
				time_now,
				task,
				experiment_name,
				plot_func.__name__.replace("_", "-"),
			)

		n_simulations = self.specification.num_simulations
		levels, coverages = expected_coverage_mc(
			posterior=self.estimator.flow,
			pairs=((theta, x) for theta, x in islice(self.loader, n_simulations)),
		)

		fig = coverage_plot(levels, coverages, legend=task)
		self.present_fig(
			fig,
			outdir,
			time_now,
			task,
			experiment_name,
			coverage_plot.__name__.replace("_", "-"),
		)

		trace_df = pd.DataFrame(
			posterior_samples.cpu().detach().numpy(), columns=self.names
		)

		for name in trace_df:
			estimate = trace_df[name].mean()
			uncertainty = trace_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trace_df, "trace")



"""Contains the implementations for likelihood-free methods using
ELFI

Implements the supported likelihood-free methods using
the ELFI library.

"""

import time

import elfi
import numpy as np
from matplotlib import pyplot as plt

from ...data_model import ParameterEstimateModel
from ..base import ELFIBase


class ELFILikelihoodFree(ELFIBase):
	"""The ELFI likelihood-free method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""

		def simulator_func(
			*parameter_values: list,
			size: int | None = None,
			batch_size: int = 1,
			random_state: int | None = None,
		) -> np.ndarray:
			parameters = {}
			for i, name in enumerate(self.names):
				parameters[name] = parameter_values[i].item()

			abc_kwargs = self.get_calibration_func_kwargs()
			simulation_id = self.get_simulation_uuid()
			observed_data = self.specification.observed_data
			results = self.call_calibration_func(
				parameters, simulation_id, observed_data, **abc_kwargs
			)
			if not hasattr(results, "__iter__"):
				results = [results]
			if not isinstance(results, np.ndarray):
				results = np.array(results)

			return results

		self.create_simulator(simulator_func, take_log=False)

		sampler_name = self.specification.method
		supported_samplers = dict(
			bolfi=elfi.BOLFI,
		)
		sampler_class = supported_samplers.get(sampler_name, None)
		if sampler_class is None:
			raise ValueError(
				f"Unsupported ELFI sampler: {sampler_name}.",
				f"Supported ELFI samplers are {', '.join(supported_samplers)}",
			)

		n_init = self.specification.n_init
		sampler_kwargs = dict(
			target_name=self.distance,
			initial_evidence=n_init,
			batch_size=1,
			update_interval=n_init // 10,
			acq_noise_var=self.specification.acq_noise_var,
			bounds=self.bounds,
			seed=self.specification.random_seed,
		)

		self.sampler = sampler_class(self.distance.model, **sampler_kwargs)

		n_iterations = self.specification.n_iterations
		objective_kwargs = dict(n_evidence=n_iterations)
		self.sampler.set_objective(**objective_kwargs)

		walltime = self.specification.walltime
		walltime = time.time() + walltime * 60
		while not self.sampler.finished and time.time() < walltime:
			self.sampler.iterate()

		n_samples = self.specification.n_samples
		n_chains = self.specification.n_chains
		sampler = self.specification.sampler
		epsilon = self.specification.epsilon

		self.history = self.sampler.sample(
			n_samples, algorithm=sampler, n_chains=n_chains, threshold=epsilon
		)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		for plot_func in [
			self.sampler.plot_state,
			self.sampler.plot_discrepancy,
			self.sampler.plot_gp,
			self.history.plot_traces,
			self.history.plot_marginals,
			self.history.plot_pairs,
		]:
			plot_func(figsize=self.specification.figsize)
			if outdir is not None:
				plot_name = plot_func.__name__.replace("_", "-")
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-{plot_name}.png",
				)
				self.append_artifact(outfile)
				plt.tight_layout()
				plt.savefig(outfile)
				plt.close()
			else:
				plt.show()
				plt.close()

		for name, values in self.history.samples.items():
			estimate = values.mean()  # Unweighted samples
			uncertainty = values.std()
			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		outfile = self.join(
			outdir,
			f"{time_now}-{task}-{experiment_name}-trace.csv",
		)
		self.append_artifact(outfile)
		self.history.save(outfile)



"""Contains the implementations for simulation-based inference methods using
SBI

Implements the supported simulation-based inference methods using
the SBI library.

"""

import numpy as np
import pandas as pd
import torch.nn as nn
from matplotlib import pyplot as plt
from sbi import analysis as analysis
from sbi import utils as utils
from sbi.inference import (
	SNPE,
	prepare_for_sbi,
	simulate_for_sbi,
)

from ..base import SimulationBasedInferenceBase
from ..data_model import ParameterEstimateModel


class SBISimulationBasedInference(SimulationBasedInferenceBase):
	"""The SBI simulation-based inference method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""

		sbi_kwargs = self.get_calibration_func_kwargs()

		def simulator_func(X: np.ndarray) -> np.ndarray:
			X = X.detach().cpu().numpy()
			X = [X]
			results = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				sbi_kwargs,
			)
			return results[0]

		simulator, prior = prepare_for_sbi(simulator_func, self.parameters)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		embedding_net = nn.Identity()
		neural_posterior = utils.posterior_nn(
			model=self.specification.method,
			embedding_net=embedding_net,
			**method_kwargs,
		)

		inference = SNPE(prior=prior, density_estimator=neural_posterior)

		theta = self.specification.X
		x = self.specification.Y
		if theta is None or x is None:
			theta, x = simulate_for_sbi(
				simulator,
				proposal=prior,
				num_simulations=self.specification.num_simulations,
			)

		inference = inference.append_simulations(theta, x)
		density_estimator = inference.train(
			max_num_epochs=self.specification.n_iterations
		)
		posterior = inference.build_posterior(density_estimator)
		posterior.set_default_x(self.specification.observed_data)

		self.prior = prior
		self.simulator = simulator
		self.inference = inference
		self.posterior = posterior

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		n_draws = self.specification.n_samples
		posterior_samples = self.posterior.sample(
			(n_draws,), x=self.specification.observed_data
		)

		limits = []
		lower_limits, _ = posterior_samples.min(axis=0)
		upper_limits, _ = posterior_samples.max(axis=0)
		for i in range(len(self.names)):
			limits.append((lower_limits[i], upper_limits[i]))

		for plot_func in [analysis.pairplot, analysis.marginal_plot]:
			plt.rcParams.update({"font.size": 8})
			fig, _ = plot_func(
				posterior_samples,
				figsize=self.specification.figsize,
				labels=self.names,
				limits=limits,
			)
			self.present_fig(
				fig,
				outdir,
				time_now,
				task,
				experiment_name,
				plot_func.__name__.replace("_", "-"),
			)

		for plot_func in [
			analysis.conditional_pairplot,
			analysis.conditional_marginal_plot,
		]:
			plt.rcParams.update({"font.size": 8})
			fig, _ = plot_func(
				density=self.posterior,
				condition=self.posterior.sample((1,)),
				figsize=self.specification.figsize,
				labels=self.names,
				limits=limits,
			)
			self.present_fig(
				fig,
				outdir,
				time_now,
				task,
				experiment_name,
				plot_func.__name__.replace("_", "-"),
			)

		thetas = self.prior.sample((n_draws,))
		xs = self.simulator(thetas)
		ranks, dap_samples = analysis.run_sbc(
			thetas, xs, self.posterior, num_posterior_samples=n_draws
		)

		num_bins = None
		if n_draws <= 20:
			num_bins = n_draws

		for plot_type in ["hist", "cdf"]:
			plt.rcParams.update({"font.size": 8})
			fig, _ = analysis.sbc_rank_plot(
				ranks=ranks,
				num_bins=num_bins,
				num_posterior_samples=n_draws,
				plot_type=plot_type,
				parameter_labels=self.names,
			)
			fig_suffix = (
				f"{analysis.sbc_rank_plot.__name__.replace('_', '-')}_{plot_type}"
			)
			self.present_fig(fig, outdir, time_now, task, experiment_name, fig_suffix)

		trace_df = pd.DataFrame(
			posterior_samples.cpu().detach().numpy(), columns=self.names
		)
		for name in trace_df:
			estimate = trace_df[name].mean()
			uncertainty = trace_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trace_df, "trace")

		check_stats = analysis.check_sbc(
			ranks, thetas, dap_samples, num_posterior_samples=n_draws
		)

		check_stats_list = []
		for metric in check_stats:
			metric_dict = {"metric": metric}
			check_stats_list.append(metric_dict)
			scores = check_stats[metric].detach().cpu().numpy()
			for i, score in enumerate(scores):
				col_name = self.names[i]
				metric_dict[col_name] = score

		check_stats_df = pd.DataFrame(check_stats_list)
		self.to_csv(check_stats_df, "diagnostics")



"""Contains the implementations for uncertainty analysis methods using
Chaospy

Implements the supported uncertainty analysis methods using
the Chaospy library.

"""

import chaospy
import gstools
import numpy as np
import sklearn.linear_model as lm
from matplotlib import pyplot as plt

from ..base import CalibrationWorkflowBase


class ChaospyUncertaintyAnalysis(CalibrationWorkflowBase):
	"""The Chaospy uncertainty analysis method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		parameters = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			data_type = spec.data_type
			self.data_types.append(data_type)

			distribution_name = (
				spec.distribution_name.replace("_", " ").title().replace(" ", "")
			)
			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			dist_instance = getattr(chaospy, distribution_name)
			parameter = dist_instance(*distribution_args, **distribution_kwargs)
			parameters.append(parameter)

		self.parameters = chaospy.J(*parameters)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		solvers = ["linear", "gp", "quadrature"]
		solver_name = self.specification.solver
		if solver_name not in solvers:
			raise ValueError(
				f"Unsupported solver: {solver_name}.",
				f"Supported solvers are {', '.join(solvers)}",
			)

		order = self.specification.order
		rule = self.specification.method
		X = self.specification.X
		if X is None:
			if solver_name == "quadrature":
				nodes, weights = chaospy.generate_quadrature(
					order, self.parameters, rule=rule
				)
				X = nodes.T
			else:
				n_samples = self.specification.n_samples
				X = self.parameters.sample(n_samples, rule=rule).T

		n_replicates = self.specification.n_replicates
		if n_replicates > 1:
			X = np.repeat(X, n_replicates, axis=0)
			self.rng.shuffle(X)

		uncertainty_kwargs = self.get_calibration_func_kwargs()
		Y = self.specification.Y
		if Y is None:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				uncertainty_kwargs,
			)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		if solver_name == "gp":
			method_kwargs["normed"] = True

		expansion = chaospy.generate_expansion(order, self.parameters, **method_kwargs)

		linear_kwargs = {"fit_intercept": False}
		linear_models = dict(
			least_squares=lm.LinearRegression(**linear_kwargs),
			elastic=lm.ElasticNet(alpha=0.5, **linear_kwargs),
			lasso=lm.Lasso(**linear_kwargs),
			lasso_lars=lm.LassoLars(**linear_kwargs),
			lars=lm.Lars(**linear_kwargs),
			ridge=lm.Ridge(**linear_kwargs),
		)
		linear_regression = self.specification.algorithm
		model = linear_models.get(linear_regression, None)

		if solver_name == "quadrature":
			model_approx = chaospy.fit_quadrature(
				expansion,
				nodes,  # type: ignore[possibly-undefined]
				weights,  # type: ignore[possibly-undefined]
				Y,
			)
		else:
			model_approx = chaospy.fit_regression(
				expansion,
				X.T,
				Y,
				model=model,
				retall=False,
			)

		if solver_name == "gp":
			if (
				self.specification.flatten_Y
				and len(Y.shape) > 1
				and self.specification.X is None
			):
				X = self.extend_X(X, Y.shape[1])
				Y = Y.flatten()

			gp = gstools.Gaussian(dim=X.shape[-1])
			self.krige = gstools.krige.Universal(gp, X.T, Y, list(expansion))
			self.krige(X)

		self.emulator = model_approx

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		solver_name = self.specification.solver

		expected = chaospy.E(self.emulator, self.parameters)
		std = chaospy.Std(self.emulator, self.parameters)

		fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]
		observed_data = self.specification.observed_data
		X = np.arange(0, expected.shape[0], 1)
		axes[0].plot(X, observed_data)
		axes[0].set_title(f"Observed {output_label}")

		axes[1].plot(X, expected)
		axes[1].set_title(f"Emulated {output_label} for {solver_name} solver")
		axes[1].fill_between(X, expected - std, expected + std, alpha=0.5)
		self.present_fig(fig, outdir, time_now, task, experiment_name, "emulated")

		if solver_name == "gp":
			mu, sigma = self.krige.field, np.sqrt(self.krige.krige_var)

			obs_shape = observed_data.shape[0]
			mu_shape = mu.shape[0]
			if obs_shape != mu_shape:
				mu = mu.reshape(obs_shape, int(mu_shape / obs_shape)).mean(axis=1)
				sigma = sigma.reshape(obs_shape, int(mu_shape / obs_shape)).mean(axis=1)

			fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)
			axes[0].plot(X, observed_data)
			axes[0].set_title(f"Observed {output_label}")
			axes[1].plot(X, mu)
			axes[1].set_title(f"Emulated {output_label} for Polynomial Kriging")
			axes[1].fill_between(X, mu - sigma, mu + sigma, alpha=0.5)
			self.present_fig(
				fig, outdir, time_now, task, experiment_name, "polynomial-kriging"
			)



"""Contains the implementations for Approximate Bayesian Computation methods using
PyABC

Implements the supported Approximate Bayesian Computation methods using
the PyABC library.

"""

from datetime import timedelta

import numpy as np
import pandas as pd
import pyabc
from matplotlib import pyplot as plt

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterDataType, ParameterEstimateModel


class PyABCApproximateBayesianComputation(CalibrationWorkflowBase):
	"""The PyABC Approximate Bayesian Computation method class."""

	def dist_name_processing(self, name: str) -> str:
		"""Apply data preprocessing to the distribution name.

		Args:
			name (str): The unprocessed distribution name.

		Returns:
			str: The processed distribution name.
		"""
		name = name.replace(" ", "_").lower()

		if name == "normal":
			name = "norm"
		return name

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		distributions = {}
		transition_mapping = {}
		parameter_spec = self.specification.parameter_spec.parameters

		for spec in parameter_spec:
			parameter_name = spec.name
			data_type = spec.data_type
			if data_type == ParameterDataType.DISCRETE:
				lower_bound, upper_bound = self.get_parameter_bounds(spec)
				lower_bound = np.floor(lower_bound).astype("int")
				upper_bound = np.floor(upper_bound).astype("int")
				discrete_domain = np.arange(lower_bound, upper_bound + 1)

				distributions[parameter_name] = pyabc.RV(
					"rv_discrete",
					values=(
						discrete_domain,
						np.repeat(1 / len(discrete_domain), len(discrete_domain)),
					),
				)
				transition_mapping[parameter_name] = pyabc.DiscreteJumpTransition(
					domain=discrete_domain, p_stay=0.7
				)
			else:
				distribution_name = self.dist_name_processing(spec.distribution_name)
				distribution_args = spec.distribution_args
				if distribution_args is None:
					distribution_args = []

				distribution_kwargs = spec.distribution_kwargs
				if distribution_kwargs is None:
					distribution_kwargs = {}

				distributions[parameter_name] = pyabc.RV(
					distribution_name, *distribution_args, **distribution_kwargs
				)
				transition_mapping[parameter_name] = pyabc.MultivariateNormalTransition(
					scaling=1
				)

		self.parameters = pyabc.Distribution(**distributions)
		self.transitions = pyabc.AggregatedTransition(mapping=transition_mapping)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		adaptive_pop_size = pyabc.AdaptivePopulationSize(
			self.specification.n_init,
			self.specification.min_population_size,
			self.specification.n_bootstrap,
		)

		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]

		abc_kwargs = self.get_calibration_func_kwargs()

		def simulator_func(parameters: dict) -> dict:
			observed_data = self.specification.observed_data
			simulation_id = self.get_simulation_uuid()

			results = self.call_calibration_func(
				parameters, simulation_id, observed_data, **abc_kwargs
			)

			summary_stats = {}
			if len(output_labels) == 1:
				summary_stats[output_labels[0]] = results
			else:
				for i, output_label in enumerate(output_labels):
					summary_stats[output_label] = results[i]  # type: ignore[index]

			return summary_stats

		n_jobs = self.specification.n_jobs
		if n_jobs > 1:
			sampler = pyabc.MulticoreEvalParallelSampler(n_procs=n_jobs)
		else:
			sampler = pyabc.SingleCoreSampler(check_max_eval=True)

		distance_func = pyabc.AggregatedDistance(
			[
				lambda simulated, _: simulated[output_label]
				for output_label in output_labels
			]
		)

		self.abc = pyabc.ABCSMC(
			simulator_func,
			self.parameters,
			distance_func,
			population_size=adaptive_pop_size,
			transitions=self.transitions,
			eps=pyabc.MedianEpsilon(),
			sampler=sampler,
		)

		storage = self.specification.storage
		if storage is None:
			storage = "sqlite://"
		self.abc.new(storage)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		self.history = self.abc.run(
			minimum_epsilon=self.specification.epsilon,
			max_walltime=timedelta(minutes=self.specification.walltime),
			**method_kwargs,
		)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		for plot_func in [
			pyabc.visualization.plot_sample_numbers,
			pyabc.visualization.plot_total_sample_numbers,
			pyabc.visualization.plot_sample_numbers_trajectory,
			pyabc.visualization.plot_epsilons,
			pyabc.visualization.plot_effective_sample_sizes,
			pyabc.visualization.plot_walltime,
			pyabc.visualization.plot_total_walltime,
			pyabc.visualization.plot_contour_matrix,
			pyabc.visualization.plot_acceptance_rates_trajectory,
			pyabc.visualization.plot_kde_matrix_highlevel,
		]:
			plot_func(self.history)
			plot_name = plot_func.__name__.replace("_", "-")
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-{plot_name}.png",
				)
				self.append_artifact(outfile)
				plt.tight_layout()
				plt.savefig(outfile)
				plt.close()
			else:
				plt.show()
				plt.close()

		distribution_dfs = []
		for t in range(self.history.max_t + 1):
			df, w = self.history.get_distribution(m=0, t=t)
			df["t"] = t
			df["w"] = w
			distribution_dfs.append(df)
		distribution_dfs = pd.concat(distribution_dfs)

		max_t = distribution_dfs["t"].max()
		trace_df = distribution_dfs.query(f"t == {max_t}")
		trace_df = trace_df.drop(columns=["t", "w"])

		for name in trace_df:
			estimate = trace_df[name].mean()
			uncertainty = trace_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(distribution_dfs, "parameters")

		populations_df = self.history.get_all_populations()
		self.to_csv(populations_df, "populations")

		population_particles_df = self.history.get_nr_particles_per_population()
		self.to_csv(population_particles_df, "particles-per-population")

		distances_df = []
		for t in range(self.history.max_t + 1):
			df = self.history.get_weighted_distances(t=t)
			df["t"] = t
			distances_df.append(df)
		distances_df = pd.concat(distances_df)
		self.to_csv(distances_df, "distances")



"""Contains the implementations for uncertainty analysis methods using
Pygpc

Implements the supported uncertainty analysis methods using
the Pygpc library.

"""

from collections import OrderedDict
from collections.abc import Callable

import numpy as np
import pygpc
from matplotlib import pyplot as plt
from pygpc.AbstractModel import AbstractModel

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterDataType


class PygpcModel(AbstractModel):
	def __init__(
		self,
		workflow: CalibrationWorkflowBase,
		parameter_names: list[str],
		data_types: list[str],
	):
		"""PygpcModel constructor.

		Args:
			workflow (CalibrationWorkflowBase): The calibration workflow.
			parameter_names (list[str]): The simulation parameter names.
			data_types (list[str]): The parameter data types.
		"""
		super(type(self), self).__init__(matlab_model=False)
		self.call_calibration_func = workflow.call_calibration_func
		self.observed_data = workflow.specification.observed_data
		self.batched = workflow.specification.batched
		self.uncertainty_kwargs = workflow.get_calibration_func_kwargs()
		self.workflow = workflow

		self.parameter_names = parameter_names
		self.data_types = data_types

	def validate(self) -> None:
		"""Validate the simulation data."""
		pass

	def simulate(
		self, process_id: int | None = None, matlab_engine: Callable | None = None
	) -> np.ndarray:
		"""Run the simulation.

		Args:
			process_id (int | None, optional): The simulation process ID.
				Defaults to None.
			matlab_engine (Callable | None, optional): The MATLAB Python engine.
				Defaults to None.

		Returns:
			np.ndarray: The simulation output data.
		"""
		parameter_name = self.parameter_names[0]
		N = self.p[parameter_name].shape[0]

		parameters = []
		for i in range(N):
			parameter_set = {}
			for j, parameter_name in enumerate(self.parameter_names):
				parameter_value = self.p[parameter_name][i]
				data_type = self.data_types[j]
				if data_type == ParameterDataType.CONTINUOUS:  # type: ignore[comparison-overlap]
					parameter_set[parameter_name] = parameter_value
				else:
					parameter_set[parameter_name] = int(parameter_value)
			parameters.append(parameter_set)

		simulation_ids = [
			self.workflow.get_simulation_uuid() for _ in range(len(parameters))
		]

		if self.batched:
			results = self.call_calibration_func(
				parameters,
				simulation_ids,
				self.observed_data,
				**self.uncertainty_kwargs,
			)
		else:
			results = []
			for i, parameter in enumerate(parameters):
				simulation_id = simulation_ids[i]
				result = self.call_calibration_func(
					parameter,
					simulation_id,
					self.observed_data,
					**self.uncertainty_kwargs,
				)
				results.append(result)  # type: ignore[arg-type]
		results = np.array(results)

		if len(results.shape) == 1:
			results = results[:, np.newaxis]
		return results


class PygpcUncertaintyAnalysis(CalibrationWorkflowBase):
	"""The Pygpc uncertainty analysis method class."""

	def dist_name_processing(self, name: str) -> str:
		"""Apply data preprocessing to the distribution name.

		Args:
		    name (str): The unprocessed distribution name.

		Returns:
		    str: The processed distribution name.
		"""
		name = name.replace("_", " ").title().replace(" ", "")

		if name == "Normal":
			name = "Norm"
		return name

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		names = []
		data_types = []
		parameters = OrderedDict()

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			names.append(parameter_name)

			data_type = spec.data_type
			data_types.append(data_type)

			distribution_name = self.dist_name_processing(spec.distribution_name)
			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			if len(distribution_kwargs.keys()) == 0 and len(distribution_args) == 2:
				distribution_args = [distribution_args]

			dist_instance = getattr(pygpc, distribution_name)
			parameter = dist_instance(*distribution_args, **distribution_kwargs)
			parameters[parameter_name] = parameter

		self.parameters = parameters
		self.model = PygpcModel(self, names, data_types)  # type: ignore[arg-type]
		self.problem = pygpc.Problem(self.model, parameters)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		_, time_now, experiment_name, outdir = self.prepare_analyze()
		if outdir is None:
			fn_results = None
		else:
			experiment_name = (
				self.specification.experiment_name  # type: ignore[assignment]
			)
			fn_results = self.join(outdir, f"{time_now}_{experiment_name}")

		options = self.specification.method_kwargs
		if options is None:
			options = {}

		options["method"] = self.specification.method
		options["solver"] = self.specification.solver
		options["n_cpu"] = self.specification.n_jobs
		options["n_grid"] = self.specification.n_init
		options["grid_options"] = dict(seed=self.specification.random_seed)
		options["grid"] = pygpc.Random
		options["fn_results"] = fn_results
		options["verbose"] = self.specification.verbose

		algorithm_name = self.specification.algorithm
		algorithms = dict(
			static_io=pygpc.Static_IO,
			static=pygpc.Static,
			me_static=pygpc.MEStatic,
			me_static_io=pygpc.MEStatic_IO,
			static_projection=pygpc.StaticProjection,
			me_static_projection=pygpc.MEStaticProjection,
			reg_adaptive=pygpc.RegAdaptive,
			me_reg_adaptive_projection=pygpc.MERegAdaptiveProjection,
			reg_adaptive_projection=pygpc.RegAdaptiveProjection,
		)
		algorithm_class = algorithms.get(algorithm_name, None)
		if algorithm_class is None:
			raise ValueError(
				f"Unsupported Pygpc algorithm: {algorithm_name}.",
				f"Supported Pygpc algorithm are {', '.join(algorithms)}",
			)
		algorithm = algorithm_class(problem=self.problem, options=options)

		session = pygpc.Session(algorithm=algorithm)
		session, coeffs, results = session.run()

		self.session = session
		self.coeffs = coeffs
		self.results = results
		self.options = options

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		n_samples = self.specification.n_samples
		outfile = None

		output_label = self.specification.output_labels[0]  # type: ignore[index]
		observed_data = self.specification.observed_data
		X = np.arange(0, observed_data.shape[-1], 1)
		fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)
		axes[0].plot(X, observed_data)
		axes[0].set_title(f"Observed {output_label}")

		for i in range(self.results.shape[0]):
			axes[1].plot(X, self.results[i])
		axes[1].set_title(f"Emulated {output_label}")
		self.present_fig(
			fig, outdir, time_now, task, experiment_name, f"emulated-{output_label}"
		)

		plot_func = pygpc.validate_gpc_mc
		if outdir is not None:
			plot_name = plot_func.__name__.replace("_", "-")
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-{plot_name}",
			)
			self.append_artifact(outfile)
		plot_func(
			session=self.session,
			coeffs=self.coeffs,
			fn_out=outfile,
			n_cpu=self.session.n_cpu,
		)
		if outdir is None:
			plt.show()
			plt.close()

		plot_func = pygpc.validate_gpc_plot
		if outdir is not None:
			plot_name = plot_func.__name__.replace("_", "-")
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-{plot_name}",
			)
			self.append_artifact(outfile)
		plot_func(
			session=self.session,
			coeffs=self.coeffs,
			random_vars=self.model.parameter_names,
			fn_out=outfile,
			n_grid=[n_samples, n_samples],
			n_cpu=self.session.n_cpu,
		)
		if outdir is None:
			plt.show()
			plt.close()



"""Contains the implementations for optimisation methods using Emukit

Implements the supported optimisation methods using the Emukit library.

"""

import numpy as np
import pandas as pd
from emukit.bayesian_optimization.acquisitions import (
	ExpectedImprovement,
	NegativeLowerConfidenceBound,
	ProbabilityOfImprovement,
)
from emukit.bayesian_optimization.acquisitions.local_penalization import (
	LocalPenalization,
)
from emukit.bayesian_optimization.loops import BayesianOptimizationLoop
from emukit.core.initial_designs import RandomDesign
from matplotlib import pyplot as plt

from ..base import EmukitBase
from ..data_model import ParameterEstimateModel
from ..estimators import EmukitEstimator


class EmukitOptimisation(EmukitBase):
	"""The Emukit optimisation method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		objective_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			return self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				objective_kwargs,
				True,
			)

		n_init = self.specification.n_init
		X, Y = self.get_X_Y(n_init, target_function)
		method_kwargs = self.specification.method_kwargs
		estimator = EmukitEstimator(method_kwargs)
		estimator.fit(X, Y)

		acquisition_name = self.specification.acquisition_func
		acquisition_funcs = dict(
			ei=ExpectedImprovement,
			poi=ProbabilityOfImprovement,
			lp=LocalPenalization,
			nlcb=NegativeLowerConfidenceBound,
		)
		acquisition_func = acquisition_funcs.get(acquisition_name, None)
		if acquisition_func is None:
			raise ValueError(
				f"Unsupported acquisition function: {acquisition_name}.",
				f"Supported acquisition functions are {', '.join(acquisition_funcs)}",
			)
		acquisition = acquisition_func(model=estimator.emulator)

		optimisation_loop = BayesianOptimizationLoop(
			model=estimator.emulator,
			space=self.parameter_space,
			acquisition=acquisition,
			batch_size=1,
		)

		n_iterations = self.specification.n_iterations
		optimisation_loop.run_loop(target_function, n_iterations)
		self.emulator = estimator
		self.optimisation_loop = optimisation_loop

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		results = self.optimisation_loop.get_results()
		self.results = results
		trial_history = results.best_found_value_per_iteration

		fig, ax = plt.subplots(figsize=self.specification.figsize)
		t = np.arange(0, len(trial_history), 1)
		ax.plot(t, trial_history)
		ax.set_title("Optimisation history")
		self.present_fig(
			fig, outdir, time_now, task, experiment_name, "plot-optimization-history"
		)

		optimised_parameters = results.minimum_location
		parameter_dict = {}
		for i, name in enumerate(self.names):
			parameter_dict[name] = optimised_parameters[i]
		parameter_df = pd.DataFrame(parameter_dict, index=[0])

		for name in parameter_df.columns:
			estimate = parameter_df[name].item()
			parameter_estimate = ParameterEstimateModel(name=name, estimate=estimate)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(parameter_df, "parameters")

		if self.specification.use_shap and outdir is not None:
			design = RandomDesign(self.parameter_space)
			n_samples = self.specification.n_samples
			X_sample = design.get_samples(n_samples)
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-param-importances.png",
			)
			self.calculate_shap_importances(
				X_sample,
				self.emulator,
				self.names,
				self.specification.test_size,
				outfile,
			)



"""Contains the implementations for uncertainty analysis methods using
OpenTurns

Implements the supported uncertainty analysis methods using
the OpenTurns library.

"""

import numpy as np
import openturns as ot
import openturns.viewer as viewer

from ..base import OpenTurnsBase
from ..estimators import FunctionalChaosEstimator, KrigingEstimator


class OpenTurnsUncertaintyAnalysis(OpenTurnsBase):
	"""The OpenTurns uncertainty analysis method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		uncertainty_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				uncertainty_kwargs,
			)
			if len(Y.shape) == 1:
				Y = np.expand_dims(Y, axis=1)
			return Y

		n_samples = self.specification.n_samples
		X, Y = self.get_X_Y(n_samples, target_function)

		test_size = self.specification.test_size
		if test_size > 0:
			test_indx = int(test_size * len(X))
			self.X_test = X[:test_indx, :]
			self.Y_test = Y[:test_indx, :]

			X = X[test_indx:, :]
			Y = Y[test_indx:, :]
		else:
			self.X_test = None
			self.Y_test = None

		solvers = ["functional_chaos", "kriging"]
		solver_name = self.specification.solver
		if solver_name not in solvers:
			raise ValueError(
				f"Unsupported solver: {solver_name}.",
				f"Supported solvers are {', '.join(solvers)}",
			)

		if solver_name == "functional_chaos":
			estimator = FunctionalChaosEstimator(
				self.parameters, self.specification.order
			)
		else:
			method_kwargs = self.specification.method_kwargs
			if method_kwargs is None:
				method_kwargs = {}

			method_kwargs["covariance_scale"] = method_kwargs.get(
				"covariance_scale", 1.0
			)
			method_kwargs["covariance_amplitude"] = method_kwargs.get(
				"covariance_amplitude", 1.0
			)
			method_kwargs["parameters"] = self.parameters
			method_kwargs["n_out"] = self.specification.n_out
			estimator = KrigingEstimator(**method_kwargs)

		estimator.fit(X, Y)
		self.emulator = estimator

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		solver_name = self.specification.solver
		input_dim = self.parameters.getDimension()

		solver_name = self.specification.solver
		if solver_name == "functional_chaos":
			sobol_indices = ot.FunctionalChaosSobolIndices(self.emulator.result)

			first_order = [sobol_indices.getSobolIndex(i) for i in range(input_dim)]
			total_order = [
				sobol_indices.getSobolTotalIndex(i) for i in range(input_dim)
			]
			graph = ot.SobolIndicesAlgorithm.DrawSobolIndices(
				self.names, first_order, total_order
			)

			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			if outdir is not None:
				outfile = self.join(
					outdir, f"{time_now}-{task}-{experiment_name}-sobol-indices.png"
				)
				self.append_artifact(outfile)
				view.save(outfile)

		if self.X_test is not None and self.Y_test is not None:
			y_pred = self.emulator.predict(self.X_test)
			val = ot.MetaModelValidation(self.Y_test, y_pred)
			graph = val.drawValidation()
			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			r2 = self.emulator.score(self.X_test, self.Y_test)
			graph.setTitle(f"R2: {r2}")

			if outdir is not None:
				outfile = self.join(
					outdir, f"{time_now}-{task}-{experiment_name}-r2.png"
				)
				self.append_artifact(outfile)
				view.save(outfile)



"""Contains the implementations for surrogate modelling methods using GPyTorch

Implements the supported surrogate modelling methods using the GPyTorch library.

"""

import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
from sklearn.preprocessing import MinMaxScaler

from ..base import SurrogateBase
from ..estimators import get_single_task_exact_gp


class GPyTorchSurrogateModel(SurrogateBase):
	"""The GPyTorch surrogate modelling method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		surrogate_kwargs = self.get_calibration_func_kwargs()
		n_samples = self.specification.n_samples

		X = self.specification.X
		if X is None:
			X = self.sample_parameters(n_samples)

		Y = self.specification.Y
		if Y is None:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				surrogate_kwargs,
			)

		self.Y_shape = Y.shape
		if self.specification.flatten_Y and len(self.Y_shape) > 1:
			X = self.extend_X(X, self.Y_shape[1])
			Y = Y.flatten()

		X_scaler = MinMaxScaler()
		X_scaler.fit(X)
		self.X_scaler = X_scaler

		if torch.cuda.is_available():
			device = "cuda"
		else:
			device = "cpu"

		X = torch.tensor(X_scaler.transform(X), dtype=torch.double)
		Y = torch.tensor(Y, dtype=torch.double)

		model = get_single_task_exact_gp(
			lr=self.specification.lr,
			max_epochs=self.specification.n_iterations,
			device=device,
		)
		model.fit(X, Y)

		self.emulator = model
		self.device = device
		self.X = X
		self.Y = Y

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		model = self.emulator
		X = self.X
		Y = self.Y

		n_samples = self.specification.n_samples
		X_sample = self.parameters.sample(n_samples, rule="sobol").T
		if self.specification.flatten_Y and len(self.Y_shape) > 1:
			X_sample = self.extend_X(X_sample, self.Y_shape[1])
		X_sample = torch.tensor(self.X_scaler.transform(X_sample), dtype=torch.double)
		Y_sample = model.predict(X_sample, return_std=False)

		names = self.names.copy()
		if X.shape[1] > len(names):
			names.append("_dummy_index")
		df = pd.DataFrame(X, columns=names)

		if self.specification.use_shap and outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-param-importances.png",
			)
			self.calculate_shap_importances(
				df, self.emulator, names, self.specification.test_size, outfile
			)

		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]
		if len(self.Y_shape) == 1:
			df[f"simulated_{output_label}"] = Y
			fig, axes = plt.subplots(
				nrows=len(self.names), figsize=self.specification.figsize
			)
			if not isinstance(axes, np.ndarray):
				axes = [axes]

			for i, parameter_name in enumerate(self.names):
				df.plot.scatter(
					parameter_name,
					f"simulated_{output_label}",
					ax=axes[i],
					title=f"simulated_{output_label} against {parameter_name}",
				)
			self.present_fig(fig, outdir, time_now, task, experiment_name, "plot-slice")

			fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)
			df = pd.DataFrame(
				{
					"index": np.arange(0, Y.shape[0], 1),
					"simulated": Y,
					"emulated": Y_sample,
				}
			)
			df.plot.scatter(
				"index", "simulated", ax=axes[0], title=f"Simulated {output_label}"
			)
			df.plot.scatter(
				"index", "emulated", ax=axes[1], title=f"Emulated {output_label}"
			)
			self.present_fig(
				fig, outdir, time_now, task, experiment_name, f"emulated-{output_label}"
			)
		else:
			if self.specification.flatten_Y:
				df[f"simulated_{output_label}"] = Y
				fig, axes = plt.subplots(
					nrows=len(self.names), figsize=self.specification.figsize
				)
				if not isinstance(axes, np.ndarray):
					axes = [axes]
				for i, parameter_name in enumerate(self.names):
					df.plot.scatter(
						parameter_name,
						f"simulated_{output_label}",
						ax=axes[i],
						title=f"simulated_{output_label} against {parameter_name}",
					)
				self.present_fig(
					fig, outdir, time_now, task, experiment_name, "plot-slice"
				)

				reshaped_Y_sample = Y_sample.reshape(self.Y_shape)
				Y = self.Y.reshape(self.Y_shape).detach().cpu().numpy()
				index = np.arange(0, reshaped_Y_sample.shape[1], 1)

				fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)

				for i in range(Y.shape[0]):
					axes[0].plot(index, Y[i])
				axes[0].set_title(f"Simulated {output_label}")

				for i in range(reshaped_Y_sample.shape[0]):
					axes[1].plot(index, reshaped_Y_sample[i])
				axes[1].set_title(f"Emulated {output_label}")
				self.present_fig(
					fig,
					outdir,
					time_now,
					task,
					experiment_name,
					f"emulated-{output_label}",
				)



"""Contains an implementation of a Lorenz95 model.

An implementation of a Lorenz95 example model.

"""

import numpy as np
import pandas as pd

from ..base import ExampleModelBase


class Lorenz95(ExampleModelBase):
	"""The Lorenz95 model."""

	def __init__(self) -> None:
		"""Lorenz95 constructor."""
		super().__init__()

		self.GROUND_TRUTH = dict(F=5, steps=1000, dt=0.01, N=40, noise_std=0)

	def lorenz95(self, x: np.ndarray, F: float) -> np.ndarray:
		"""The Lorenz 95 model.

		Args:
		    x (np.ndarray): The input data.
		    F (float): The force.

		Returns:
		    np.ndarray: The simulation output.
		"""
		N = len(x)
		dxdt = []
		for i in range(N):
			y = (x[(i + 1) % N] - x[i - 2]) * x[i - 1] - x[i] + F
			dxdt.append(y)
		return np.array(dxdt)

	def rk4_step(self, x: np.ndarray, F: float, dt: float) -> np.ndarray:
		"""Runge-Kutta 4th order integration.

		Args:
		    x (np.ndarray): The input data.
		    F (float): The force.
		    dt (float): Rate of change over time.

		Returns:
		    np.ndarray: The integrated values.
		"""
		k1 = self.lorenz95(x, F)
		k2 = self.lorenz95(x + 0.5 * dt * k1, F)
		k3 = self.lorenz95(x + 0.5 * dt * k2, F)
		k4 = self.lorenz95(x + dt * k3, F)
		return x + (dt / 6) * (k1 + 2 * k2 + 2 * k3 + k4)

	def get_observed_data(self) -> np.ndarray | pd.DataFrame:
		"""Retrieve observed data.

		Returns:
		    np.ndarray | pd.DataFrame: The observed data.
		"""
		observed_df = self.simulate(self.GROUND_TRUTH)
		return observed_df

	def simulate(self, parameters: dict) -> np.ndarray | pd.DataFrame:
		"""Run the simulation.

		Args:
		    parameters (dict): The simulation parameters.

		Returns:
		    np.ndarray | pd.DataFrame: The simulated data.
		"""
		N = parameters["N"]
		F = parameters["F"]
		steps = parameters["steps"]

		x = F * np.ones(N)
		x[0] += 1e-7
		trajectory = [x.copy()]

		for _ in range(steps):
			x = self.step(parameters, x)
			trajectory.append(x.copy())

		trajectory = np.array(trajectory)
		df = pd.DataFrame(trajectory)
		return df

	def step(self, parameters: dict, x: np.ndarray | None = None) -> np.ndarray:
		dt = parameters["dt"]
		F = parameters["F"]
		N = parameters["N"]
		noise_std = np.exp(parameters["noise_std"])

		if x is None:
			x = F * np.ones(N)
			x[0] += 1e-7

		rng = np.random.default_rng()
		noise = rng.normal(0, noise_std, size=N)
		x = self.rk4_step(x, F, dt) + noise
		return x



"""Contains the simulation-based inference base class

The defined base class for performing simulation-based inference.

"""

import numpy as np
import torch
import torch.distributions as dist

from ..data_model import ParameterDataType
from .calibration_base import CalibrationWorkflowBase


class SimulationBasedInferenceBase(CalibrationWorkflowBase):
	"""The simulation-based inference base class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		self.parameters = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			name = spec.name
			self.names.append(name)

			data_type = spec.data_type
			self.data_types.append(data_type)

			if data_type == ParameterDataType.DISCRETE:
				lower_bound, upper_bound = self.get_parameter_bounds(spec)
				lower_bound = np.floor(lower_bound).astype("int")
				upper_bound = np.floor(upper_bound).astype("int")
				replicates = np.floor(upper_bound - lower_bound).astype("int")
				probabilities = torch.tensor([1 / replicates])
				probabilities = probabilities.repeat(replicates)
				base_distribution = dist.Categorical(probabilities)
				transforms = [
					dist.AffineTransform(
						loc=torch.Tensor([lower_bound]), scale=torch.Tensor([1])
					)
				]
				prior = dist.TransformedDistribution(base_distribution, transforms)
			else:
				distribution_name = (
					spec.distribution_name.replace("_", " ").title().replace(" ", "")
				)

				distribution_args = spec.distribution_args
				if distribution_args is None:
					distribution_args = []
				distribution_args = [torch.Tensor([arg]) for arg in distribution_args]

				distribution_kwargs = spec.distribution_kwargs
				if distribution_kwargs is None:
					distribution_kwargs = {}
				distribution_kwargs = {
					k: torch.Tensor([v]) for k, v in distribution_kwargs.items()
				}

				distribution_class = getattr(dist, distribution_name)

				prior = distribution_class(*distribution_args, **distribution_kwargs)

			self.parameters.append(prior)



"""Contains the implementations for the surrogate modelling methods

Implements the supported surrogate modelling methods.

"""

import importlib
from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .sklearn_wrapper import SklearnSurrogateModel

TASK = "surrogate_modelling"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	sklearn=SklearnSurrogateModel
)

if importlib.util.find_spec("gpytorch") is not None:
	from .gpytorch_wrapper import GPyTorchSurrogateModel

	IMPLEMENTATIONS["gpytorch"] = GPyTorchSurrogateModel


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for surrogate modelling.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for surrogate modelling.
	"""
	return IMPLEMENTATIONS


class SurrogateModelMethodModel(CalibrationModel):
	"""The surrogate modelling method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	batch_size: int = Field(
		description="The batch size when training the surrogate model", default=1000
	)
	flatten_Y: bool = Field(description="Flatten the simulation outputs", default=False)


class SurrogateModelMethod(CalibrationMethodBase):
	"""The surrogate modelling method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: SurrogateModelMethodModel,
		engine: str = "sklearn",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""SurrogateModelMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (SurrogateModelMethodModel): The calibration
				specification.
		    engine (str, optional): The surrogate modelling backend.
				Defaults to "sklearn".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains an implementation of the Lotka Volterra model.

An implementation of a Lotka Volterra example model.

"""

import numpy as np
import pandas as pd
from scipy.integrate import solve_ivp

from ..base import ExampleModelBase

# Refer to https://www.pymc.io/projects/examples/en/latest/samplers/SMC-ABC_Lotka-Volterra_example.html


class LotkaVolterraModel(ExampleModelBase):
	"""Lotka Volterra simulation model."""

	def __init__(self) -> None:
		"""LotkaVolterraModel constructor."""
		super().__init__()
		self.GROUND_TRUTH = dict(
			alpha=0.52,
			beta=0.024,
			h0=34.0,
			l0=5.9,
			t=np.arange(1900.0, 1921.0, 1),
			gamma=0.84,
			delta=0.026,
		)
		self.OUTPUT_LABELS = ["lynx", "hare"]

	def get_observed_data(self) -> np.ndarray | pd.DataFrame:
		"""Retrieve observed data.

		Returns:
		    np.ndarray | pd.DataFrame: The observed data.
		"""
		observed_df = pd.DataFrame(
			dict(
				year=np.arange(1900.0, 1921.0, 1),
				lynx=np.array(
					[
						4.0,
						6.1,
						9.8,
						35.2,
						59.4,
						41.7,
						19.0,
						13.0,
						8.3,
						9.1,
						7.4,
						8.0,
						12.3,
						19.5,
						45.7,
						51.1,
						29.7,
						15.8,
						9.7,
						10.1,
						8.6,
					]
				),
				hare=np.array(
					[
						30.0,
						47.2,
						70.2,
						77.4,
						36.3,
						20.6,
						18.1,
						21.4,
						22.0,
						25.4,
						27.1,
						40.3,
						57.0,
						76.6,
						52.3,
						19.5,
						11.2,
						7.6,
						14.6,
						16.2,
						24.7,
					]
				),
			)
		)
		return observed_df

	def simulate(self, parameters: dict) -> np.ndarray | pd.DataFrame:
		"""Run the simulation.

		Args:
			parameters (dict): The simulation parameters.

		Returns:
		    np.ndarray | pd.DataFrame: The simulated data.
		"""

		def dX_dt(_: np.ndarray, X: np.ndarray) -> np.ndarray:
			x, y = X
			dx_dt = parameters["alpha"] * x - parameters["beta"] * x * y
			dy_dt = -parameters["gamma"] * y + parameters["delta"] * x * y
			return np.array([dx_dt, dy_dt])

		X0 = [parameters["h0"], parameters["l0"]]
		t = (parameters["t"].min(), parameters["t"].max())
		x_y = solve_ivp(fun=dX_dt, y0=X0, t_span=t, t_eval=parameters["t"].values).y

		df = pd.DataFrame(dict(lynx=x_y[1, :], hare=x_y[0, :]))
		return df



"""Contains base classes for the various calibration methods

Abstract base classes are defined for the
simulation calibration procedures.

"""

import os.path as osp
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps
from typing import Any

import numpy as np
import pandas as pd
import shap
import uncertainty_toolbox as uct
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from shap import KernelExplainer
from sklearn.base import BaseEstimator

from ..data_model import (
	CalibrationModel,
	DistributionModel,
	ParameterDataType,
	ParameterEstimateModel,
	ParameterEstimatesModel,
	ParameterSpecification,
)
from ..statistics import get_full_factorial_design
from ..utils import (
	calibration_func_wrapper,
	create_file_path,
	extend_X,
	get_datetime_now,
	get_simulation_uuid,
)


def pre_post_hooks(f: Callable) -> Callable:
	"""Execute prehooks and posthooks for calibration methods.

	Args:
	    f (Callable): The wrapped function.

	Returns:
	    Callable: The wrapper function.
	"""

	@wraps(f)
	def wrapper(
		self: CalibrationWorkflowBase, *args: list, **kwargs: dict
	) -> "CalibrationWorkflowBase":
		"""The wrapper function for prehooks and posthooks.

		Returns:
		    CalibrationWorkflowBase: The calibration workflow.
		"""
		func_name = f.__name__
		getattr(self, f"prehook_{func_name}")()
		result = f(self, *args, **kwargs)
		getattr(self, f"posthook_{func_name}")()
		return result

	return wrapper


class CalibrationWorkflowBase(ABC):
	"""The calibration workflow abstract class."""

	def __init__(
		self, calibration_func: Callable, specification: CalibrationModel, task: str
	) -> None:
		"""CalibrationMethodBase constructor.

		Args:
		    calibration_func (Callable): The calibration function.
		        For example, a simulation function or objective function.
		    specification (CalibrationModel): The calibration specification.
		    task (str): The calibration task.
		"""
		super().__init__()
		self.task = task
		self.calibration_func = calibration_func
		self.specification = specification
		self.artifacts: list[str] = []
		self.parameter_estimates = ParameterEstimatesModel(estimates=[])

		random_seed = self.specification.random_seed
		self.rng = np.random.default_rng(random_seed)

	@abstractmethod
	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure.

		Raises:
		    NotImplementedError: Error raised for the unimplemented abstract method.
		"""
		pass

	@abstractmethod
	def execute(self) -> None:
		"""Execute the simulation calibration procedure.

		Raises:
		    NotImplementedError: Error raised for the unimplemented abstract method.
		"""
		pass

	@abstractmethod
	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure.

		Raises:
		    NotImplementedError: Error raised for the unimplemented abstract method.
		"""
		pass

	def prehook_specify(self) -> None:
		"""Prehook to run before specify()."""
		pass

	def posthook_specify(self) -> None:
		"""Posthook to run after specify()."""
		pass

	def prehook_execute(self) -> None:
		"""Prehook to run before execute()."""
		pass

	def posthook_execute(self) -> None:
		"""Posthook to run after execute()."""
		pass

	def prehook_analyze(self) -> None:
		"""Prehook to run before analyze()."""
		pass

	def posthook_analyze(self) -> None:
		"""Posthook to run after analyze()."""
		pass

	def prepare_analyze(self) -> tuple[str, str, str, str | None]:
		"""Perform preparations for the analyze step.

		Returns:
		    tuple[str, str, str,  str | None]: A list of
		        metadata needed for the analysis outputs.
		"""
		task = self.task
		time_now = get_datetime_now()
		experiment_name = self.specification.experiment_name
		self.time_now = time_now
		outdir = self.specification.outdir
		return task, time_now, experiment_name, outdir  # type: ignore[return-value]

	def get_simulation_uuid(self) -> str:
		"""Get a new simulation uuid.

		Returns:
		    str: The simulation uuid.
		"""
		return get_simulation_uuid()

	def extend_X(self, X: np.ndarray, Y_rows: int) -> np.ndarray:
		"""Extend the number of rows for X with a dummy index column.

		Args:
		    X (np.ndarray): The input matrix.
		    Y_rows (int) The number of rows for the simulation outputs.

		Returns:
		    np.ndarray: The extended input matrix with a dummy column.
		"""
		return extend_X(X, Y_rows)

	def get_default_rng(self, random_seed: int | None = None) -> np.random.Generator:
		"""Get a numpy random number generator.

		Args:
		    random_seed (int | None, optional): The
		        random seed. Defaults to None.

		Returns:
		    np.random.Generator: The random number generator.
		"""
		return np.random.default_rng(random_seed)

	def join(self, *paths: str) -> str:
		"""Join file paths.

		Args:
		    paths (str): The file paths.

		Returns:
		    str: The joined file paths.
		"""
		return osp.join(*paths)

	def create_file_path(self, file_path: str) -> str:
		"""Create file path if it does not exist.

		Args:
		    file_path (str): The file path to create.

		Returns:
		    str: The created file path.
		"""
		return create_file_path(file_path)

	def get_parameter_bounds(self, spec: DistributionModel) -> tuple[float, float]:
		"""Get the lower and upper bounds from a parameter specification.

		Args:
		    spec (DistributionModel): The parameter specification.

		Raises:
		    ValueError: Error raised when the
		        bounds cannot be identified.

		Returns:
		    tuple[float, float]: The lower and upper bounds.
		"""
		distribution_args = spec.distribution_args
		if isinstance(distribution_args, list):
			if len(distribution_args) == 2:
				lower_bound, upper_bound = distribution_args
				return lower_bound, upper_bound

		distribution_kwargs = spec.distribution_kwargs
		if isinstance(distribution_kwargs, dict):
			lower_bound = distribution_kwargs.get("lower_bound", None)
			upper_bound = distribution_kwargs.get("upper_bound", None)
			if lower_bound is not None and upper_bound is not None:
				return lower_bound, upper_bound

		raise ValueError(f"Invalid parameter specification for {spec.name}")

	def get_calibration_func_kwargs(self) -> dict:
		"""Get the calibration function named arguments.

		Returns:
		    dict: The calibration function named arguments.
		"""
		calibration_func_kwargs = self.specification.calibration_func_kwargs
		if calibration_func_kwargs is None:
			calibration_func_kwargs = {}

		pass_calibration_workflow = self.specification.pass_calibration_workflow
		if pass_calibration_workflow is not None:
			k = "calibration_workflow"
			if isinstance(pass_calibration_workflow, str):
				k = pass_calibration_workflow
			calibration_func_kwargs[k] = self

		return calibration_func_kwargs

	def prehook_calibration_func(
		self,
		parameters: dict | list[dict],
		simulation_id: str | list[str],
		observed_data: np.ndarray | None,
		**method_kwargs: dict,
	) -> tuple:
		"""Prehook to run before calling the calibration function

		Args:
		    parameters (dict | List[dict]): The simulation parameters.
		    simulation_id (str | List[str]): The simulation IDs.
		    observed_data (np.ndarray | None): The observed data.

		Returns:
		    tuple: The calibration function parameters.
		"""
		return parameters, simulation_id, observed_data, method_kwargs

	def posthook_calibration_func(
		self,
		results: np.ndarray | pd.DataFrame | float,
		parameters: dict | list[dict],
		simulation_id: str | list[str],
		observed_data: np.ndarray | None,
		**method_kwargs: dict,
	) -> tuple:
		"""Posthook to run after calling the calibration function

		Args:
		    results (np.ndarray | pd.DataFrame | float): The simulation
		        results.
		    parameters (dict | List[dict]): The simulation parameters.
		    simulation_id (str | List[str]): The simulation IDs.
		    observed_data (np.ndarray | None): The observed data.

		Returns:
		    tuple: The calibration function results and parameters.
		"""
		return results, parameters, simulation_id, observed_data, method_kwargs

	def call_calibration_func(
		self,
		parameters: dict | list[dict],
		simulation_id: str | list[str],
		observed_data: np.ndarray | None,
		**method_kwargs: dict,
	) -> float | list[float] | np.ndarray | pd.DataFrame:
		"""Wrapper method for the calibration function.

		Args:
		    results (np.ndarray | pd.DataFrame | float): The simulation
		        results.
		    parameters (dict | List[dict]): The simulation parameters.
		    simulation_id (str | List[str]): The simulation IDs.
		    observed_data (np.ndarray | None): The observed data.

		Returns:
		    float | list[float] | np.ndarray | pd.DataFrame: The
		        calibration function results.
		"""
		prehook_results = self.prehook_calibration_func(
			parameters, simulation_id, observed_data, **method_kwargs
		)
		parameters, simulation_id, observed_data, method_kwargs = prehook_results

		results = self.calibration_func(
			parameters, simulation_id, observed_data, **method_kwargs
		)

		results, *_ = self.posthook_calibration_func(
			results, parameters, simulation_id, observed_data, **method_kwargs
		)
		return results

	def calibration_func_wrapper(
		self,
		X: np.ndarray,
		workflow: "CalibrationWorkflowBase",
		observed_data: pd.DataFrame | np.ndarray,
		parameter_names: list[str],
		data_types: list[ParameterDataType],
		calibration_kwargs: dict,
		wrap_values: bool = False,
	) -> np.ndarray:
		"""Wrapper function for the calibration function.

		Args:
		    X (np.ndarray): The parameter set matrix.
		    workflow (CalibrationWorkflowBase): The calibration workflow.
		    observed_data (pd.DataFrame | np.ndarray): The observed data.
		    parameter_names (list[str]): The list of simulation parameter names.
		    data_types (list[ParameterDataType]): The data types for each parameter.
		    calibration_kwargs (dict): Arguments to supply to the calibration function.
		    wrap_values (bool): Whether to wrap scalar values with a list.
		        Defaults to False.

		Returns:
		    np.ndarray: The simulation output data.
		"""
		return calibration_func_wrapper(
			X,
			workflow,
			observed_data,
			parameter_names,
			data_types,
			calibration_kwargs,
			wrap_values,
		)

	def get_full_factorial_design(
		self, parameter_spec: ParameterSpecification | None = None
	) -> np.ndarray:
		"""Get a full factorial design from a parameter specification.

		Args:
		    parameter_spec (ParameterSpecification | None, optional):
		        The simulation parameter specification. Defaults to None.

		Returns:
		    np.ndarray: The full factorial design.
		"""
		if parameter_spec is None:
			parameter_spec = self.specification.parameter_spec
		return get_full_factorial_design(parameter_spec)  # type: ignore[arg-type]

	def get_artifacts(self) -> list[str]:
		"""Getter method for the artifact list.

		Returns:
		    list[str]: The calibration workflow artifact list.
		"""
		return self.artifacts

	def append_artifact(self, artifact: str) -> None:
		"""Add a new artifact to the artifacts list.

		Args:
		    artifact (str): The artifact to append.
		"""
		self.artifacts.append(artifact)

	def present_fig(
		self,
		fig: Figure,
		outdir: str | None,
		time_now: str,
		task: str,
		experiment_name: str,
		suffix: str,
	) -> None:
		"""Present the figure by showing or writing to file.

		Args:
		    fig (Figure): The matplotlib figure.
		    outdir (str | None): The image output directory.
		    time_now (str): The current time.
		    task (str): The current calibration task.
		    suffix (str): The file name suffix.
		"""
		fig.tight_layout()
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-{suffix}.png"
			)
			self.append_artifact(outfile)
			fig.savefig(outfile)
		else:
			fig.show()

	def plot_simulated_vs_observed(
		self,
		simulated_data: np.ndarray,
		observed_data: np.ndarray,
		outdir: str,
		time_now: str,
		task: str,
		experiment_name: str,
		label: str = "",
	) -> None:
		"""Plot simulated data against observed data.

		Args:
		    simulated_data (np.ndarray): The simulated data.
		    observed_data (np.ndarray): The observed data.
		    outdir (str): The output directory.
		    time_now (str): The current time.
		    task (str): The calibration task.
		    experiment_name (str): The experiment name.
		    label (str, optional): The plot axes label. Defaults to "".
		"""
		simulated_label = "simulated"
		observed_label = "observed"

		if label != "":
			simulated_label += f" {label}"
			observed_label += f" {label}"

		df = pd.DataFrame(
			{simulated_label: simulated_data, observed_label: observed_data}
		)
		df["index"] = df.index

		fig, axes = plt.subplots(nrows=3, figsize=self.specification.figsize)
		df.plot.scatter("index", simulated_label, ax=axes[0])
		df.plot.scatter("index", observed_label, ax=axes[1])
		df.plot.scatter(simulated_label, observed_label, ax=axes[2])

		plot_suffix = f"{simulated_label}_vs_{observed_label}".replace(" ", "_")
		self.present_fig(fig, outdir, time_now, task, experiment_name, plot_suffix)

	def set_output_labels_from_Y(self, Y: np.ndarray) -> None:
		"""Set the simulation output labels from output data.

		Args:
		    Y (np.ndarray): The simulation outputs.
		"""
		output_labels = self.specification.output_labels

		if output_labels is None:
			if Y.ndim > 1:
				output_labels = [f"target_{i}" for i in range(Y.shape[1])]
			else:
				output_labels = ["target"]

		self.specification.output_labels = output_labels

	def get_parameter_estimates(self) -> ParameterEstimatesModel:
		"""Get the estimated parameter values, and potentially their uncertainties.

		Returns:
		    ParameterEstimatesModel: The estimated parameter values.
		"""
		return self.parameter_estimates

	def add_parameter_estimate(self, estimate: ParameterEstimateModel) -> None:
		"""Add a parameter estimate to the set of estimates.

		Args:
		    estimate (ParameterEstimateModel): The parameter estimate.
		"""
		self.parameter_estimates.estimates.append(estimate)

	def calculate_shap_importances(
		self,
		X: np.ndarray,
		emulator: BaseEstimator,
		names: list[str],
		test_size: float = 0,
		outfile: str | None = None,
	) -> None:
		"""Calculate SHAP importances using Kernel SHAP.

		Args:
		    X (np.ndarray): The training data.
		    emulator (BaseEstimator): The surrogate model.
		    names (list[str]): The parameter names.
		    test_size (float, optional): The test dataset size. Defaults to 0.
		    outfile (str | None, optional): The output file. Defaults to None.
		"""
		if test_size == 0:
			test_indx = 25
		else:
			test_indx = int(test_size * len(X))

		X_train = X[:test_indx]
		X_test = X[-test_indx:]

		explainer = KernelExplainer(emulator.predict, data=X_train, feature_names=names)
		shap_values = explainer.shap_values(X_test)

		show = False
		if outfile is None:
			show = True
		shap.summary_plot(shap_values, X_test, show=show, feature_names=names)

		if not show:
			self.append_artifact(outfile)  # type: ignore[arg-type]
			plt.tight_layout()
			plt.savefig(outfile)
			plt.close()

	def calc_uncertainty_calibration_metric(
		self,
		metric: str,
		mu: np.ndarray,
		sigma: np.ndarray,
		Y: np.ndarray,
		recal_model: BaseEstimator | None = None,
	) -> float:
		"""Calculate predictive uncertainty calibration metrics.

		Args:
		    metric (str): The metric name.
		    mu (np.ndarray): The conditional mean predictions.
		    sigma (np.ndarray): The conditional predicted standard deviations.
		    Y (np.ndarray): The simulation output data.
		    recal_model (BaseEstimator | None, optional): The prediction
		        recalibrator. Defaults to None.

		Returns:
		    float: The uncertainty calibration metric.
		"""

		metric_func = getattr(uct, metric)
		score = metric_func(mu, sigma, Y, recal_model=recal_model)
		return score

	def fit_recalibrator(
		self, emulator: BaseEstimator, mu: np.ndarray, sigma: np.ndarray, y: np.ndarray
	) -> None:
		"""Fit a model recalibrator using Isotonic regression.

		Args:
		    emulator (BaseEstimator): The surrogate model.
		    mu (np.ndarray): The conditional mean predictions.
		    sigma (np.ndarray): The conditional predicted standard deviations.
		    Y (np.ndarray): The simulation output data.
		"""
		exp_props, obs_props = uct.get_proportion_lists_vectorized(mu, sigma, y)
		recal_model = uct.iso_recal(exp_props, obs_props)
		emulator.recal_model = recal_model

	def to_csv(self, df: pd.DataFrame, file_suffix: str) -> None:
		"""Convert dataframe to csv file.

		Args:
			df (pd.DataFrame): The dataframe.
			file_suffix (str): The file name suffix.
		"""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		outfile = self.join(
			outdir,  # type: ignore[arg-type]
			f"{time_now}-{task}-{experiment_name}-{file_suffix}.csv",
		)
		self.append_artifact(outfile)
		df.to_csv(outfile, index=False)

	def get_emulator(self) -> BaseEstimator:
		"""Get the trained emulator model.

		Returns:
			BaseEstimator: The emulator.
		"""
		return self.emulator

	def get_sampler(self) -> Any:
		"""Get the parameter sampler.

		Returns:
			Any: The sampler.
		"""
		return self.sampler

	def sample_parameters(self, n_samples: int) -> np.ndarray:
		"""Sample from the parameter space.

		Args:
			n_samples (int): The number of samples.

		Returns:
			np.ndarray: The sampled parameter values.
		"""
		raise NotImplementedError("Calibrator has not implemented sample_parameters()")


class CalibrationMethodBase(CalibrationWorkflowBase):
	"""The calibration method abstract class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: CalibrationModel,
		task: str,
		engine: str,
		implementations: dict[str, type[CalibrationWorkflowBase]],
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""CalibrationMethodBase constructor.

		Args:
		    calibration_func (Callable): The calibration function.
		        For example, a simulation function or objective function.
		    specification (CalibrationModel): The calibration specification.
		    task (str): The calibration task.
		    engine (str): The calibration implementation engine.
		    implementations (dict[str, type[CalibrationWorkflowBase]]): The
		        list of supported engines.
		    implementation (CalibrationWorkflowBase | None): The
		        calibration workflow implementation.
		"""
		super().__init__(calibration_func, specification, task)
		self.engine = engine
		self.supported_engines = list(implementations.keys())

		if implementation is None:
			if engine not in self.supported_engines:
				raise NotImplementedError(f"Unsupported {task} engine: {engine}")

			implementation_class = implementations.get(engine, None)
			if implementation_class is None:
				raise ValueError(
					f"{self.task} implementation not defined for: {engine}.",
					f"Supported engines are {', '.join(self.supported_engines)}",
				)
			self.implementation = implementation_class(
				calibration_func, specification, task
			)
		else:
			self.implementation = implementation

	def _implementation_check(self, function_name: str) -> None:
		"""Check that the implementation is set.

		Args:
		    function_name (str): The name of the function.

		Raises:
		    ValueError: Error raised when the implementation is not set.
		"""
		if self.implementation is None:
			raise ValueError(
				f"{self.task} implementation is not set when calling {function_name}()."
			)

	@pre_post_hooks
	def specify(self) -> "CalibrationMethodBase":
		"""Specify the parameters of the model calibration procedure.

		Raises:
		    ValueError: Error raised when the implementation is not set.

		Returns:
		    CalibrationMethodBase: The calibration method.
		"""
		self._implementation_check("specify")
		self.implementation.specify()
		return self

	@pre_post_hooks
	def execute(self) -> "CalibrationMethodBase":
		"""Execute the simulation calibration procedure.

		Raises:
		    ValueError: Error raised when the implementation is not set.

		Returns:
		    CalibrationMethodBase: The calibration method.
		"""
		self._implementation_check("execute")
		self.implementation.execute()
		return self

	@pre_post_hooks
	def analyze(self) -> "CalibrationMethodBase":
		"""Analyze the results of the simulation calibration procedure.

		Raises:
		    ValueError: Error raised when the implementation is not set.

		Returns:
		    CalibrationMethodBase: The calibration method.
		"""
		self._implementation_check("analyze")
		self.implementation.analyze()
		return self

	def get_engines(self, as_string: bool = False) -> list | str:
		"""Get a list of supported engines.

		Args:
		    as_string (bool, optional): Whether to return
		        the engine list as a string. Defaults to False.

		Returns:
		    list | str: The list of supported engines.
		"""
		if as_string:
			return ", ".join(self.supported_engines)
		else:
			return self.supported_engines

	def get_artifacts(self) -> list[str]:
		"""Getter method for the artifact list.

		Returns:
		    list[str]: The calibration workflow artifact list.
		"""
		return self.implementation.get_artifacts()

	def get_parameter_estimates(self) -> ParameterEstimatesModel:
		"""Get the estimated parameter values, and potentially their uncertainties.

		Returns:
		    ParameterEstimatesModel: The estimated parameter values.
		"""
		self._implementation_check("get_parameter_estimates")
		return self.implementation.get_parameter_estimates()

	def get_emulator(self) -> BaseEstimator:
		"""Get the trained emulator model.

		Returns:
			BaseEstimator: The emulator.
		"""
		self._implementation_check("get_emulator")
		return self.implementation.get_emulator()

	def get_sampler(self) -> Any:
		"""Get the parameter sampler.

		Returns:
			Any: The sampler.
		"""
		self._implementation_check("get_sampler")
		return self.implementation.get_sampler()

	def sample_parameters(self, n_samples: int) -> np.ndarray:
		"""Sample from the parameter space.

		Args:
			n_samples (int): The number of samples.

		Returns:
			np.ndarray: The sampled parameter values.
		"""
		self._implementation_check("get_sampler")
		return self.implementation.sample_parameters(n_samples)



import importlib

from .emukit_estimator import EmukitEstimator
from .openturns_estimator import FunctionalChaosEstimator, KrigingEstimator

__all__ = [
	EmukitEstimator,
	FunctionalChaosEstimator,
	KrigingEstimator,
]

if importlib.util.find_spec("gpytorch") is not None:
	from .gpytorch_estimator import (
		SingleTaskGPRegressionModel,
		get_single_task_exact_gp,
	)

	__all__.extend([SingleTaskGPRegressionModel, get_single_task_exact_gp])  # type: ignore[list-item]



"""Contains the implementations for evolutionary algorithms using EvoTorch

Implements the supported evolutionary algorithms using the EvoTorch library.

"""

from typing import Any

import evotorch.algorithms as algos
import evotorch.operators as operators
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import torch
from evotorch import Problem
from evotorch.logging import PandasLogger
from matplotlib import pyplot as plt
from plotly.subplots import make_subplots

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterDataType, ParameterEstimateModel


class EvoTorchEvolutionary(CalibrationWorkflowBase):
	"""The EvoTorch evolutionary algorithm method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		self.bounds = []
		self.lower_bounds = []
		self.upper_bounds = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			data_type = spec.data_type
			self.data_types.append(data_type)

			bounds = self.get_parameter_bounds(spec)
			self.bounds.append(bounds)

			lower_bound, upper_bound = bounds
			if data_type == ParameterDataType.DISCRETE:
				lower_bound = np.floor(lower_bound).astype("int")
				upper_bound = np.floor(upper_bound).astype("int")
			self.lower_bounds.append(lower_bound)
			self.upper_bounds.append(upper_bound)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		directions = []
		for direction in self.specification.directions:
			if direction == "minimize":
				direction = "min"
			elif direction == "maximize":
				direction = "max"
			directions.append(direction)

		self.trials: list[dict[str, Any]] = []
		evolutionary_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: torch.Tensor) -> torch.Tensor:
			trial = {}
			if X.ndim == 1:
				for i, name in enumerate(self.names):
					trial[name] = X[i].item()
					self.trials.append(trial)
				X = [X]

			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				evolutionary_kwargs,
				False,
			)

			self.set_output_labels_from_Y(Y)
			output_labels = self.specification.output_labels
			for i, label in enumerate(output_labels):  # type: ignore[arg-type]
				trial[label] = Y[i]
			return Y

		vectorized = self.specification.batched
		seed = self.specification.random_seed
		self.problem = Problem(
			directions,
			target_function,
			solution_length=len(self.names),
			bounds=(self.lower_bounds, self.upper_bounds),
			vectorized=vectorized,
			seed=seed,
		)

		algorithm_name = self.specification.method
		supported_algorithms = dict(
			ga=algos.GeneticAlgorithm, cosyne=algos.Cosyne, cmaes=algos.PyCMAES
		)
		algorithm_class = supported_algorithms.get(algorithm_name, None)
		if algorithm_class is None:
			raise ValueError(
				f"Unsupported EvoTorch algorithm: {algorithm_name}.",
				f"Supported EvoTorch algorithms are {', '.join(supported_algorithms)}",
			)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		if algorithm_name == "cmaes":
			method_kwargs["obj_index"] = 0
		elif algorithm_name == "ga":
			operator_dict = self.specification.operators
			operator_list = []
			if operator_dict is not None:
				for operator_name, operator_kwargs in operator_dict.items():
					operator_class = getattr(operators, operator_name)
					operator = operator_class(self.problem, **operator_kwargs)
					operator_list.append(operator)
			method_kwargs["operators"] = operator_list

		self.searcher = algorithm_class(self.problem, **method_kwargs)
		interval = self.specification.n_samples
		self.logger = PandasLogger(self.searcher, interval=interval)

		num_generations = self.specification.n_iterations
		self.searcher.run(num_generations=num_generations)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		trials_df = pd.DataFrame(self.trials)
		objective_df = self.logger.to_dataframe()

		output_labels = self.specification.output_labels
		trials_df_best = trials_df.sort_values(output_labels).head(1)
		for name in self.names:
			estimate = trials_df_best[name].item()
			parameter_estimate = ParameterEstimateModel(name=name, estimate=estimate)
			self.add_parameter_estimate(parameter_estimate)

		fig, axes = plt.subplots(
			nrows=len(output_labels),  # type: ignore[arg-type]
			figsize=self.specification.figsize,
		)

		if not isinstance(axes, np.ndarray):
			axes = [axes]

		trials_df["index"] = trials_df.index
		for i, output_label in enumerate(output_labels):  # type: ignore[arg-type]
			trials_df.plot.scatter(
				"index", output_label, ax=axes[i], title=f"Simulated {output_label}"
			)
		trials_df = trials_df.drop(columns="index")

		self.present_fig(fig, outdir, time_now, task, experiment_name, "trial-history")

		parameter_names = self.names
		for output_label in output_labels:
			fig = make_subplots(
				rows=1, cols=len(parameter_names), subplot_titles=parameter_names
			)
			for i, parameter_name in enumerate(parameter_names):
				fig.add_trace(
					go.Scatter(
						x=trials_df[parameter_name],
						y=trials_df[output_label],
						mode="markers",
					),
					row=1,
					col=i + 1,
				)

			fig.update_layout(yaxis_title=output_label, showlegend=False)
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-{output_label}-plot-slice.png",
				)
				self.append_artifact(outfile)
				fig.write_image(outfile)
			else:
				fig.show()

		if outdir is None:
			return

		self.to_csv(objective_df, "objective")

		self.to_csv(trials_df, "trials")



"""Contains the implementations for Approximate Bayesian Computation methods using
ELFI

Implements the supported Approximate Bayesian Computation methods using
the ELFI library.

"""

import time

import elfi
import numpy as np
from matplotlib import pyplot as plt

from ...data_model import ParameterEstimateModel
from ..base import ELFIBase


class ELFIApproximateBayesianComputation(ELFIBase):
	"""The ELFI Approximate Bayesian Computation method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""

		def simulator_func(
			*parameter_values: list,
			size: int | None = None,
			batch_size: int = 1,
			random_state: int | None = None,
		) -> np.ndarray:
			parameters = {}
			for i, name in enumerate(self.names):
				parameters[name] = parameter_values[i].item()

			abc_kwargs = self.get_calibration_func_kwargs()
			simulation_id = self.get_simulation_uuid()
			observed_data = self.specification.observed_data
			results = self.call_calibration_func(
				parameters, simulation_id, observed_data, **abc_kwargs
			)
			if not hasattr(results, "__iter__"):
				results = [results]
			if not isinstance(results, np.ndarray):
				results = np.array(results)

			return results

		self.create_simulator(simulator_func, take_log=False)

		sampler_name = self.specification.method
		supported_samplers = dict(
			rejection=elfi.Rejection,
			smc=elfi.SMC,
			adaptive_threshold_smc=elfi.AdaptiveThresholdSMC,
		)
		sampler_class = supported_samplers.get(sampler_name, None)
		if sampler_class is None:
			raise ValueError(
				f"Unsupported ELFI sampler: {sampler_name}.",
				f"Supported ELFI samplers are {', '.join(supported_samplers)}",
			)

		sampler_kwargs = self.specification.method_kwargs
		if sampler_kwargs is None:
			sampler_kwargs = {}

		random_seed = self.specification.random_seed
		sampler_kwargs["batch_size"] = 1
		sampler_kwargs["seed"] = random_seed
		self.sampler = sampler_class(self.distance, **sampler_kwargs)

		n_samples = self.specification.n_samples
		quantile = self.specification.quantile
		n_iterations = self.specification.n_iterations
		self.sampler.bar = True
		if sampler_name == "rejection":
			self.sampler.set_objective(n_samples, quantile=quantile)
		elif sampler_name == "smc":
			quantiles = [quantile for _ in range(n_iterations)]
			self.sampler.set_objective(n_samples, quantiles=quantiles)
		elif sampler_name == "adaptive_threshold_smc":
			self.sampler.set_objective(n_samples, max_iter=n_iterations)
		else:
			raise ValueError(f"Unsupported ELFI sampler: {sampler_name}")

		walltime = self.specification.walltime
		walltime = time.time() + walltime * 60
		while not self.sampler.finished and time.time() < walltime:
			self.sampler.iterate()

		self.history = self.sampler.extract_result()

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		n_samples = self.specification.n_samples
		num_bins = 25
		if n_samples <= 20:
			num_bins = n_samples

		for plot_func in [self.history.plot_marginals, self.history.plot_pairs]:
			plot_func(bins=num_bins, figsize=self.specification.figsize)
			if outdir is not None:
				plot_name = plot_func.__name__.replace("_", "-")
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-{plot_name}.png",
				)
				self.append_artifact(outfile)
				plt.tight_layout()
				plt.savefig(outfile)
				plt.close()
			else:
				plt.show()
				plt.close()

		for name, values in self.history.samples.items():
			estimate = values.mean()  # Unweighted samples
			uncertainty = values.std()
			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		outfile = self.join(
			outdir,
			f"{time_now}-{task}-{experiment_name}-trace.csv",
		)
		self.append_artifact(outfile)
		self.history.save(outfile)



from .implementation import (
	ExperimentalDesignMethod,
	ExperimentalDesignMethodModel,
	get_implementations,
)

_all__ = [
	ExperimentalDesignMethod,
	ExperimentalDesignMethodModel,
	get_implementations,
]



"""Contains the history matching base class

The defined base class for performing history matching.

"""

import numpy as np

from .calibration_base import CalibrationWorkflowBase


class HistoryMatchingBase(CalibrationWorkflowBase):
	"""The history matching base class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		ensemble_size = self.specification.n_samples
		n_replicates = self.specification.n_replicates
		parameter_spec = self.specification.parameter_spec.parameters

		self.parameters = {}
		for spec in parameter_spec:
			parameter_name = spec.name
			distribution_name = spec.distribution_name.replace(" ", "_").lower()

			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}
			distribution_kwargs["size"] = ensemble_size

			dist_instance = getattr(self.rng, distribution_name)
			samples = dist_instance(*distribution_args, **distribution_kwargs)
			samples = np.repeat(samples, n_replicates)
			self.rng.shuffle(samples)
			self.parameters[parameter_name] = samples

		self.specification.n_samples = ensemble_size * n_replicates



"""Contains the implementations for sensitivity analysis methods using
OpenTurns

Implements the supported sensitivity analysis methods using
the OpenTurns library.

"""

import numpy as np
import openturns as ot
import openturns.viewer as viewer

from ..base import OpenTurnsBase
from ..estimators import FunctionalChaosEstimator


class OpenTurnsSensitivityAnalysis(OpenTurnsBase):
	"""The OpenTurns sensitivity analysis method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		method = self.specification.method
		sobol_methods = dict(
			saltelli=ot.SaltelliSensitivityAlgorithm,
			martinez=ot.MartinezSensitivityAlgorithm,
			jansen=ot.JansenSensitivityAlgorithm,
			mauntz_kucherenko=ot.MauntzKucherenkoSensitivityAlgorithm,
		)
		self.sobol_method_names = list(sobol_methods.keys())
		self.supported_methods = self.sobol_method_names.copy()
		self.supported_methods.extend(["chaos_sobol", "chaos_ancova"])

		if method not in self.supported_methods:
			raise ValueError(
				f"Unsupported sensitivity analysis algorithm: {method}.",
				f"Supported algorithms are: {', '.join(self.supported_methods)}",
			)

		sensitivity_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				sensitivity_kwargs,
			)
			if len(Y.shape) == 1:
				Y = np.expand_dims(Y, axis=1)
			return Y

		n_samples = self.specification.n_samples

		X = self.specification.X
		if X is None:
			sobol_experiment = ot.SobolIndicesExperiment(self.parameters, n_samples)
			X = sobol_experiment.generate()
			X = self.replicate_samples(X)
		Y = self.specification.Y
		if Y is None:
			ot_func_wrapper = self.get_ot_func_wrapper(target_function)
			Y = ot_func_wrapper(X)

		if method in self.sobol_method_names:
			self.sp = sobol_methods.get(method)(X, Y, n_samples)
			self.first_order = self.sp.getFirstOrderIndices(0)
			self.total_order = self.sp.getTotalOrderIndices(0)
			return

		estimator = FunctionalChaosEstimator(self.parameters, self.specification.order)

		estimator.fit(X, Y)
		self.emulator = estimator

		input_dim = self.parameters.getDimension()
		if method == "chaos_sobol":
			self.sp = ot.FunctionalChaosSobolIndices(self.emulator.result)
			self.first_order = [self.sp.getSobolIndex(i) for i in range(input_dim)]
			self.total_order = [self.sp.getSobolTotalIndex(i) for i in range(input_dim)]
		elif method == "chaos_ancova":
			self.sp = ot.ANCOVA(self.emulator.result, X)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		method = self.specification.method

		if method in self.sobol_method_names:
			graph = self.sp.draw()
			graph.setXTitle(" ".join(self.names))
			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-sobol-{method}-indices.png",
				)
				self.append_artifact(outfile)
				view.save(outfile)
		elif method == "chaos_sobol":
			graph = ot.SobolIndicesAlgorithm.DrawSobolIndices(
				self.names, self.first_order, self.total_order
			)
			graph.setXTitle(" ".join(self.names))
			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-chaos-sobol-indices.png",
				)
				self.append_artifact(outfile)
				view.save(outfile)
		elif method == "chaos_ancova":
			indices = self.sp.getIndices()
			uncorrelated_indices = self.sp.getUncorrelatedIndices()
			correlated_indices = indices - uncorrelated_indices

			graph = ot.SobolIndicesAlgorithm.DrawImportanceFactors(
				indices, self.parameters.getDescription(), "ANCOVA indices (Sobol)"
			)
			graph.setXTitle(" ".join(self.names))
			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			if outdir is not None:
				outfile = self.join(
					outdir, f"{time_now}-{task}-{experiment_name}-ancova-indices.png"
				)
				self.append_artifact(outfile)
				view.save(outfile)

			graph = ot.SobolIndicesAlgorithm.DrawImportanceFactors(
				uncorrelated_indices,
				self.parameters.getDescription(),
				"ANCOVA uncorrelated indices",
			)
			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-uncorrelated-ancova-indices.png",
				)
				self.append_artifact(outfile)
				view.save(outfile)

			graph = ot.SobolIndicesAlgorithm.DrawImportanceFactors(
				correlated_indices,
				self.parameters.getDescription(),
				"ANCOVA correlated indices",
			)
			view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-correlated-ancova-indices.png",
				)
				self.append_artifact(outfile)
				view.save(outfile)
		else:
			raise ValueError(
				f"Unsupported sensitivity analysis algorithm: {method}.",
				f"Supported algorithms are: {', '.join(self.supported_methods)}",
			)



"""Contains the implementations for the uncertainty analysis methods

Implements the supported uncertainty analysis methods.

"""

from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .chaospy_wrapper import ChaospyUncertaintyAnalysis
from .openturns_wrapper import OpenTurnsUncertaintyAnalysis
from .pygpc_wrapper import PygpcUncertaintyAnalysis

TASK = "uncertainty_analysis"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	chaospy=ChaospyUncertaintyAnalysis,
	pygpc=PygpcUncertaintyAnalysis,
	openturns=OpenTurnsUncertaintyAnalysis,
)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for the uncertainty analysis.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for the uncertainty analysis.
	"""
	return IMPLEMENTATIONS


class UncertaintyAnalysisMethodModel(CalibrationModel):
	"""The uncertainty analysis method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	flatten_Y: bool = Field(description="Flatten the simulation outputs", default=False)
	order: int = Field(
		description="The order for polynomial chaos expansion", default=2
	)
	solver: str | list[str] = Field(
		description="The solver for performing the uncertainty analysis",
		default="linear",
	)
	algorithm: str = Field(
		description="The algorithm for the uncertainty analysis", default=""
	)


class UncertaintyAnalysisMethod(CalibrationMethodBase):
	"""The uncertainty analysis method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: UncertaintyAnalysisMethodModel,
		engine: str = "chaospy",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""UncertaintyAnalysisMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (UncertaintyAnalysisMethodModel): The calibration
				specification.
		    engine (str, optional): The uncertainty analysis backend.
				Defaults to "chaospy".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains an implementation of the anharmonic oscillator model.

An implementation of an anharmonic oscillator example model.

"""

import numpy as np
import pandas as pd

from ..base import ExampleModelBase

# Refer to https://iterative-ensemble-smoother.readthedocs.io/en/latest/Oscillator.html


class AnharmonicOscillator(ExampleModelBase):
	"""Anharmonic oscillator simulation model."""

	def __init__(self) -> None:
		"""AnharmonicOscillator constructor."""
		super().__init__()

		self.reset_y()

		self.GROUND_TRUTH = {"omega": 3.5e-2, "lambda": 3e-4, "K": 2000}
		self.OUTPUT_LABELS = ["y"]

	def reset_y(self) -> None:
		"""Reset the trajectory vector."""
		self.y = [0, 1]

	def get_observed_data(self) -> np.ndarray | pd.DataFrame:
		"""Retrieve observed data.

		Returns:
		    np.ndarray | pd.DataFrame: The observed data.
		"""
		observed_df = self.simulate(self.GROUND_TRUTH)
		observed_df["t"] = observed_df.index
		return observed_df

	def get_df(self) -> pd.DataFrame:
		"""Get the simulation dataframe.

		Returns:
			pd.DataFrame: The simulation dataframe.
		"""
		return pd.DataFrame(dict(y=self.y))

	def simulate(self, parameters: dict) -> np.ndarray | pd.DataFrame:
		"""Run the simulation.

		Args:
			parameters (dict): The simulation parameters.

		Returns:
		    np.ndarray | pd.DataFrame: The simulated data.
		"""
		self.reset_y()
		K = parameters["K"]
		for _ in range(len(self.y), K):
			self.step(parameters)

		df = self.get_df()
		return df

	def step(self, parameters: dict) -> None:
		"""Run the model for one time step.

		Args:
			parameters (dict): The simulation parameters.
		"""
		y_k = (
			self.y[-1]
			* (
				2
				+ parameters["omega"] ** 2
				- parameters["lambda"] ** 2 * self.y[-1] ** 2
			)
			- self.y[-2]
		)

		self.y.append(y_k)
